#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, sys
from waflib import Errors, Logs, Node, Options, Task, Utils
from waflib.TaskGen import extension, before_method, after_method, feature
from waflib.Configure import conf

FRAG = '''
#include <Python.h>
#ifdef __cplusplus
extern "C" {
#endif
	void Py_Initialize(void);
	void Py_Finalize(void);
#ifdef __cplusplus
}
#endif
int main(int argc, char **argv)
{
   (void)argc; (void)argv;
   Py_Initialize();
   Py_Finalize();
   return 0;
}
'''
INST = '''
import sys, py_compile
py_compile.compile(sys.argv[1], sys.argv[2], sys.argv[3], True)
'''
DISTUTILS_IMP = ['from distutils.sysconfig import get_config_var, get_python_lib']


@before_method('process_source')
@feature('py')
def feature_py(self):
    self.install_path = getattr(self, 'install_path', '${PYTHONDIR}')
    install_from = getattr(self, 'install_from', None)
    if install_from and not isinstance(install_from, Node.Node):
        install_from = self.path.find_dir(install_from)
    self.install_from = install_from
    ver = self.env.PYTHON_VERSION
    if not ver:
        self.bld.fatal('Installing python files requires PYTHON_VERSION, try conf.check_python_version')
    if int(ver.replace('.', '')) > 31:
        self.install_32 = True


@extension('.py')
def process_py(self, node):
    assert (hasattr(self, 'install_path')
            ), 'add features="py" for target "%s" in "%s/wscript".' % (self.target, self.path.nice_path())
    self.install_from = getattr(self, 'install_from', None)
    relative_trick = getattr(self, 'relative_trick', True)
    if self.install_from:
        assert isinstance(self.install_from, Node.Node), 'add features="py" for target "%s" in "%s/wscript" (%s).' % (
            self.target, self.path.nice_path(), type(self.install_from)
        )
    if self.install_path:
        if self.install_from:
            self.add_install_files(
                install_to=self.install_path, install_from=node, cwd=self.install_from, relative_trick=relative_trick
            )
        else:
            self.add_install_files(install_to=self.install_path, install_from=node, relative_trick=relative_trick)
    lst = []
    if self.env.PYC:
        lst.append('pyc')
    if self.env.PYO:
        lst.append('pyo')
    if self.install_path:
        if self.install_from:
            target_dir = node.path_from(self.install_from) if relative_trick else node.name
            pyd = Utils.subst_vars("%s/%s" % (self.install_path, target_dir), self.env)
        else:
            target_dir = node.path_from(self.path) if relative_trick else node.name
            pyd = Utils.subst_vars("%s/%s" % (self.install_path, target_dir), self.env)
    else:
        pyd = node.abspath()
    for ext in lst:
        if self.env.PYTAG and not self.env.NOPYCACHE:
            name = node.name[:-3]
            pyobj = node.parent.get_bld().make_node('__pycache__').make_node("%s.%s.%s" % (name, self.env.PYTAG, ext))
            pyobj.parent.mkdir()
        else:
            pyobj = node.change_ext(".%s" % ext)
        tsk = self.create_task(ext, node, pyobj)
        tsk.pyd = pyd
        if self.install_path:
            self.add_install_files(
                install_to=os.path.dirname(pyd),
                install_from=pyobj,
                cwd=node.parent.get_bld(),
                relative_trick=relative_trick
            )


class pyc(Task.Task):
    color = 'PINK'

    def __str__(self):
        node = self.outputs[0]
        return node.path_from(node.ctx.launch_node())

    def run(self):
        cmd = [
            Utils.subst_vars('${PYTHON}', self.env), '-c', INST, self.inputs[0].abspath(), self.outputs[0].abspath(),
            self.pyd
        ]
        ret = self.generator.bld.exec_command(cmd)
        return ret


class pyo(Task.Task):
    color = 'PINK'

    def __str__(self):
        node = self.outputs[0]
        return node.path_from(node.ctx.launch_node())

    def run(self):
        cmd = [
            Utils.subst_vars('${PYTHON}', self.env),
            Utils.subst_vars('${PYFLAGS_OPT}', self.env), '-c', INST, self.inputs[0].abspath(),
            self.outputs[0].abspath(), self.pyd
        ]
        ret = self.generator.bld.exec_command(cmd)
        return ret


@feature('pyext')
@before_method('propagate_uselib_vars', 'apply_link')
@after_method('apply_bundle')
def init_pyext(self):
    self.uselib = self.to_list(getattr(self, 'uselib', []))
    if not 'PYEXT' in self.uselib:
        self.uselib.append('PYEXT')
    self.env.cshlib_PATTERN = self.env.cxxshlib_PATTERN = self.env.macbundle_PATTERN = self.env.pyext_PATTERN
    self.env.fcshlib_PATTERN = self.env.dshlib_PATTERN = self.env.pyext_PATTERN
    try:
        if not self.install_path:
            return
    except AttributeError:
        self.install_path = '${PYTHONARCHDIR}'


@feature('pyext')
@before_method('apply_link', 'apply_bundle')
def set_bundle(self):
    if Utils.unversioned_sys_platform() == 'darwin':
        self.mac_bundle = True


@before_method('propagate_uselib_vars')
@feature('pyembed')
def init_pyembed(self):
    self.uselib = self.to_list(getattr(self, 'uselib', []))
    if not 'PYEMBED' in self.uselib:
        self.uselib.append('PYEMBED')


@conf
def get_python_variables(self, variables, imports=None):
    if not imports:
        try:
            imports = self.python_imports
        except AttributeError:
            imports = DISTUTILS_IMP
    program = list(imports)
    program.append('')
    for v in variables:
        program.append("print(repr(%s))" % v)
    os_env = dict(os.environ)
    try:
        del os_env['MACOSX_DEPLOYMENT_TARGET']
    except KeyError:
        pass
    try:
        out = self.cmd_and_log(self.env.PYTHON + ['-c', '\n'.join(program)], env=os_env)
    except Errors.WafError:
        self.fatal('The distutils module is unusable: install "python-devel"?')
    self.to_log(out)
    return_values = []
    for s in out.splitlines():
        s = s.strip()
        if not s:
            continue
        if s == 'None':
            return_values.append(None)
        elif (s[0] == "'" and s[-1] == "'") or (s[0] == '"' and s[-1] == '"'):
            return_values.append(eval(s))
        elif s[0].isdigit():
            return_values.append(int(s))
        else:
            break
    return return_values


@conf
def test_pyembed(self, mode, msg='Testing pyembed configuration'):
    self.check(
        header_name='Python.h',
        define_name='HAVE_PYEMBED',
        msg=msg,
        fragment=FRAG,
        errmsg='Could not build a python embedded interpreter',
        features='%s %sprogram pyembed' % (mode, mode)
    )


@conf
def test_pyext(self, mode, msg='Testing pyext configuration'):
    self.check(
        header_name='Python.h',
        define_name='HAVE_PYEXT',
        msg=msg,
        fragment=FRAG,
        errmsg='Could not build python extensions',
        features='%s %sshlib pyext' % (mode, mode)
    )


@conf
def python_cross_compile(self, features='pyembed pyext'):
    features = Utils.to_list(features)
    if not (
        'PYTHON_LDFLAGS' in self.environ or 'PYTHON_PYEXT_LDFLAGS' in self.environ
        or 'PYTHON_PYEMBED_LDFLAGS' in self.environ
    ):
        return False
    for x in 'PYTHON_VERSION PYTAG pyext_PATTERN'.split():
        if not x in self.environ:
            self.fatal('Please set %s in the os environment' % x)
        else:
            self.env[x] = self.environ[x]
    xx = self.env.CXX_NAME and 'cxx' or 'c'
    if 'pyext' in features:
        flags = self.environ.get('PYTHON_PYEXT_LDFLAGS', self.environ.get('PYTHON_LDFLAGS'))
        if flags is None:
            self.fatal('No flags provided through PYTHON_PYEXT_LDFLAGS as required')
        else:
            self.parse_flags(flags, 'PYEXT')
        self.test_pyext(xx)
    if 'pyembed' in features:
        flags = self.environ.get('PYTHON_PYEMBED_LDFLAGS', self.environ.get('PYTHON_LDFLAGS'))
        if flags is None:
            self.fatal('No flags provided through PYTHON_PYEMBED_LDFLAGS as required')
        else:
            self.parse_flags(flags, 'PYEMBED')
        self.test_pyembed(xx)
    return True


@conf
def check_python_headers(conf, features='pyembed pyext'):
    features = Utils.to_list(features)
    assert ('pyembed' in features
            ) or ('pyext' in features), "check_python_headers features must include 'pyembed' and/or 'pyext'"
    env = conf.env
    if not env.CC_NAME and not env.CXX_NAME:
        conf.fatal('load a compiler first (gcc, g++, ..)')
    if conf.python_cross_compile(features):
        return
    if not env.PYTHON_VERSION:
        conf.check_python_version()
    pybin = env.PYTHON
    if not pybin:
        conf.fatal('Could not find the python executable')
    v = 'prefix SO LDFLAGS LIBDIR LIBPL INCLUDEPY Py_ENABLE_SHARED MACOSX_DEPLOYMENT_TARGET LDSHARED CFLAGS LDVERSION'.split(
    )
    try:
        lst = conf.get_python_variables(["get_config_var('%s') or ''" % x for x in v])
    except RuntimeError:
        conf.fatal("Python development headers not found (-v for details).")
    vals = ['%s = %r' % (x, y) for (x, y) in zip(v, lst)]
    conf.to_log("Configuration returned from %r:\n%s\n" % (pybin, '\n'.join(vals)))
    dct = dict(zip(v, lst))
    x = 'MACOSX_DEPLOYMENT_TARGET'
    if dct[x]:
        env[x] = conf.environ[x] = dct[x]
    env.pyext_PATTERN = '%s' + dct['SO']
    num = '.'.join(env.PYTHON_VERSION.split('.')[:2])
    conf.find_program([
        ''.join(pybin) + '-config',
        'python%s-config' % num,
        'python-config-%s' % num,
        'python%sm-config' % num
    ],
                      var='PYTHON_CONFIG',
                      msg="python-config",
                      mandatory=False)
    if env.PYTHON_CONFIG:
        if conf.env.HAVE_PYTHON_H:
            return
        all_flags = [['--cflags', '--libs', '--ldflags']]
        if sys.hexversion < 0x2070000:
            all_flags = [[k] for k in all_flags[0]]
        xx = env.CXX_NAME and 'cxx' or 'c'
        if 'pyembed' in features:
            for flags in all_flags:
                embedflags = flags + ['--embed']
                try:
                    conf.check_cfg(
                        msg='Asking python-config for pyembed %r flags' % ' '.join(embedflags),
                        path=env.PYTHON_CONFIG,
                        package='',
                        uselib_store='PYEMBED',
                        args=embedflags
                    )
                except conf.errors.ConfigurationError:
                    conf.check_cfg(
                        msg='Asking python-config for pyembed %r flags' % ' '.join(flags),
                        path=env.PYTHON_CONFIG,
                        package='',
                        uselib_store='PYEMBED',
                        args=flags
                    )
            try:
                conf.test_pyembed(xx)
            except conf.errors.ConfigurationError:
                if dct['Py_ENABLE_SHARED'] and dct['LIBDIR']:
                    env.append_unique('LIBPATH_PYEMBED', [dct['LIBDIR']])
                    conf.test_pyembed(xx)
                else:
                    raise
        if 'pyext' in features:
            for flags in all_flags:
                conf.check_cfg(
                    msg='Asking python-config for pyext %r flags' % ' '.join(flags),
                    path=env.PYTHON_CONFIG,
                    package='',
                    uselib_store='PYEXT',
                    args=flags
                )
            try:
                conf.test_pyext(xx)
            except conf.errors.ConfigurationError:
                if dct['Py_ENABLE_SHARED'] and dct['LIBDIR']:
                    env.append_unique('LIBPATH_PYEXT', [dct['LIBDIR']])
                    conf.test_pyext(xx)
                else:
                    raise
        conf.define('HAVE_PYTHON_H', 1)
        return
    all_flags = dct['LDFLAGS'] + ' ' + dct['CFLAGS']
    conf.parse_flags(all_flags, 'PYEMBED')
    all_flags = dct['LDFLAGS'] + ' ' + dct['LDSHARED'] + ' ' + dct['CFLAGS']
    conf.parse_flags(all_flags, 'PYEXT')
    result = None
    if not dct["LDVERSION"]:
        dct["LDVERSION"] = env.PYTHON_VERSION
    for name in (
        'python' + dct['LDVERSION'], 'python' + env.PYTHON_VERSION + 'm',
        'python' + env.PYTHON_VERSION.replace('.', '')
    ):
        if not result and env.LIBPATH_PYEMBED:
            path = env.LIBPATH_PYEMBED
            conf.to_log("\n\n# Trying default LIBPATH_PYEMBED: %r\n" % path)
            result = conf.check(
                lib=name,
                uselib='PYEMBED',
                libpath=path,
                mandatory=False,
                msg='Checking for library %s in LIBPATH_PYEMBED' % name
            )
        if not result and dct['LIBDIR']:
            path = [dct['LIBDIR']]
            conf.to_log("\n\n# try again with -L$python_LIBDIR: %r\n" % path)
            result = conf.check(
                lib=name,
                uselib='PYEMBED',
                libpath=path,
                mandatory=False,
                msg='Checking for library %s in LIBDIR' % name
            )
        if not result and dct['LIBPL']:
            path = [dct['LIBPL']]
            conf.to_log(
                "\n\n# try again with -L$python_LIBPL (some systems don't install the python library in $prefix/lib)\n"
            )
            result = conf.check(
                lib=name,
                uselib='PYEMBED',
                libpath=path,
                mandatory=False,
                msg='Checking for library %s in python_LIBPL' % name
            )
        if not result:
            path = [os.path.join(dct['prefix'], "libs")]
            conf.to_log("\n\n# try again with -L$prefix/libs, and pythonXY name rather than pythonX.Y (win32)\n")
            result = conf.check(
                lib=name,
                uselib='PYEMBED',
                libpath=path,
                mandatory=False,
                msg='Checking for library %s in $prefix/libs' % name
            )
        if result:
            break
    if result:
        env.LIBPATH_PYEMBED = path
        env.append_value('LIB_PYEMBED', [name])
    else:
        conf.to_log("\n\n### LIB NOT FOUND\n")
    if Utils.is_win32 or dct['Py_ENABLE_SHARED']:
        env.LIBPATH_PYEXT = env.LIBPATH_PYEMBED
        env.LIB_PYEXT = env.LIB_PYEMBED
    conf.to_log("Include path for Python extensions (found via distutils module): %r\n" % (dct['INCLUDEPY'],))
    env.INCLUDES_PYEXT = [dct['INCLUDEPY']]
    env.INCLUDES_PYEMBED = [dct['INCLUDEPY']]
    if env.CC_NAME == 'gcc':
        env.append_unique('CFLAGS_PYEMBED', ['-fno-strict-aliasing'])
        env.append_unique('CFLAGS_PYEXT', ['-fno-strict-aliasing'])
    if env.CXX_NAME == 'gcc':
        env.append_unique('CXXFLAGS_PYEMBED', ['-fno-strict-aliasing'])
        env.append_unique('CXXFLAGS_PYEXT', ['-fno-strict-aliasing'])
    if env.CC_NAME == "msvc":
        from distutils.msvccompiler import MSVCCompiler
        dist_compiler = MSVCCompiler()
        dist_compiler.initialize()
        env.append_value('CFLAGS_PYEXT', dist_compiler.compile_options)
        env.append_value('CXXFLAGS_PYEXT', dist_compiler.compile_options)
        env.append_value('LINKFLAGS_PYEXT', dist_compiler.ldflags_shared)
    conf.check(
        header_name='Python.h',
        define_name='HAVE_PYTHON_H',
        uselib='PYEMBED',
        fragment=FRAG,
        errmsg='Distutils not installed? Broken python installation? Get python-config now!'
    )


@conf
def check_python_version(conf, minver=None):
    assert minver is None or isinstance(minver, tuple)
    pybin = conf.env.PYTHON
    if not pybin:
        conf.fatal('could not find the python executable')
    cmd = pybin + ['-c', 'import sys\nfor x in sys.version_info: print(str(x))']
    Logs.debug('python: Running python command %r', cmd)
    lines = conf.cmd_and_log(cmd).split()
    assert len(lines) == 5, "found %r lines, expected 5: %r" % (len(lines), lines)
    pyver_tuple = (int(lines[0]), int(lines[1]), int(lines[2]), lines[3], int(lines[4]))
    result = (minver is None) or (pyver_tuple >= minver)
    if result:
        pyver = '.'.join([str(x) for x in pyver_tuple[:2]])
        conf.env.PYTHON_VERSION = pyver
        if 'PYTHONDIR' in conf.env:
            pydir = conf.env.PYTHONDIR
        elif 'PYTHONDIR' in conf.environ:
            pydir = conf.environ['PYTHONDIR']
        else:
            if Utils.is_win32:
                (python_LIBDEST, pydir) = conf.get_python_variables([
                    "get_config_var('LIBDEST') or ''", "get_python_lib(standard_lib=0) or ''"
                ])
            else:
                python_LIBDEST = None
                (pydir,) = conf.get_python_variables([
                    "get_python_lib(standard_lib=0, prefix=%r) or ''" % conf.env.PREFIX
                ])
            if python_LIBDEST is None:
                if conf.env.LIBDIR:
                    python_LIBDEST = os.path.join(conf.env.LIBDIR, 'python' + pyver)
                else:
                    python_LIBDEST = os.path.join(conf.env.PREFIX, 'lib', 'python' + pyver)
        if 'PYTHONARCHDIR' in conf.env:
            pyarchdir = conf.env.PYTHONARCHDIR
        elif 'PYTHONARCHDIR' in conf.environ:
            pyarchdir = conf.environ['PYTHONARCHDIR']
        else:
            (pyarchdir,) = conf.get_python_variables([
                "get_python_lib(plat_specific=1, standard_lib=0, prefix=%r) or ''" % conf.env.PREFIX
            ])
            if not pyarchdir:
                pyarchdir = pydir
        if hasattr(conf, 'define'):
            conf.define('PYTHONDIR', pydir)
            conf.define('PYTHONARCHDIR', pyarchdir)
        conf.env.PYTHONDIR = pydir
        conf.env.PYTHONARCHDIR = pyarchdir
    pyver_full = '.'.join(map(str, pyver_tuple[:3]))
    if minver is None:
        conf.msg('Checking for python version', pyver_full)
    else:
        minver_str = '.'.join(map(str, minver))
        conf.msg('Checking for python version >= %s' % (minver_str,), pyver_full, color=result and 'GREEN' or 'YELLOW')
    if not result:
        conf.fatal('The python version is too old, expecting %r' % (minver,))


PYTHON_MODULE_TEMPLATE = '''
import %s as current_module
version = getattr(current_module, '__version__', None)
if version is not None:
	print(str(version))
else:
	print('unknown version')
'''


@conf
def check_python_module(conf, module_name, condition=''):
    msg = "Checking for python module %r" % module_name
    if condition:
        msg = '%s (%s)' % (msg, condition)
    conf.start_msg(msg)
    try:
        ret = conf.cmd_and_log(conf.env.PYTHON + ['-c', PYTHON_MODULE_TEMPLATE % module_name])
    except Errors.WafError:
        conf.end_msg(False)
        conf.fatal('Could not find the python module %r' % module_name)
    ret = ret.strip()
    if condition:
        conf.end_msg(ret)
        if ret == 'unknown version':
            conf.fatal('Could not check the %s version' % module_name)
        from distutils.version import LooseVersion

        def num(*k):
            if isinstance(k[0], int):
                return LooseVersion('.'.join([str(x) for x in k]))
            else:
                return LooseVersion(k[0])

        d = {'num': num, 'ver': LooseVersion(ret)}
        ev = eval(condition, {}, d)
        if not ev:
            conf.fatal('The %s version does not satisfy the requirements' % module_name)
    else:
        if ret == 'unknown version':
            conf.end_msg(True)
        else:
            conf.end_msg(ret)


def configure(conf):
    v = conf.env
    if getattr(Options.options, 'pythondir', None):
        v.PYTHONDIR = Options.options.pythondir
    if getattr(Options.options, 'pythonarchdir', None):
        v.PYTHONARCHDIR = Options.options.pythonarchdir
    if getattr(Options.options, 'nopycache', None):
        v.NOPYCACHE = Options.options.nopycache
    if not v.PYTHON:
        v.PYTHON = [getattr(Options.options, 'python', None) or sys.executable]
    v.PYTHON = Utils.to_list(v.PYTHON)
    conf.find_program('python', var='PYTHON')
    v.PYFLAGS = ''
    v.PYFLAGS_OPT = '-O'
    v.PYC = getattr(Options.options, 'pyc', 1)
    v.PYO = getattr(Options.options, 'pyo', 1)
    try:
        v.PYTAG = conf.cmd_and_log(
            conf.env.PYTHON + [
                '-c',
                "import sys\ntry:\n print(sys.implementation.cache_tag)\nexcept AttributeError:\n import imp\n print(imp.get_tag())\n"
            ]
        ).strip()
    except Errors.WafError:
        pass


def options(opt):
    pyopt = opt.add_option_group("Python Options")
    pyopt.add_option(
        '--nopyc',
        dest='pyc',
        action='store_false',
        default=1,
        help='Do not install bytecode compiled .pyc files (configuration) [Default:install]'
    )
    pyopt.add_option(
        '--nopyo',
        dest='pyo',
        action='store_false',
        default=1,
        help='Do not install optimised compiled .pyo files (configuration) [Default:install]'
    )
    pyopt.add_option(
        '--nopycache',
        dest='nopycache',
        action='store_true',
        help='Do not use __pycache__ directory to install objects [Default:auto]'
    )
    pyopt.add_option('--python', dest="python", help='python binary to be used [Default: %s]' % sys.executable)
    pyopt.add_option(
        '--pythondir',
        dest='pythondir',
        help='Installation path for python modules (py, platform-independent .py and .pyc files)'
    )
    pyopt.add_option(
        '--pythonarchdir',
        dest='pythonarchdir',
        help='Installation path for python extension (pyext, platform-dependent .so or .dylib files)'
    )



# This hook should not be loaded at all.
# `mymodule_feature2` is imported by `mymodule_main`, but the hook for `mymodule_main` explicitly excludes
# `mymodule_feature2`. Therefore, if exclusion is applied correctly during processing of modules' imports (as opposed to
# a post-processing step), this hook will not be loaded.
raise RuntimeError("This hook should not have been loaded!")



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import sys
import pathlib

import pytest

from PyInstaller.building import utils


def test_format_binaries_and_datas_not_found_raises_error(tmp_path):
    datas = [('non-existing.txt', '.')]
    (tmp_path / 'existing.txt').touch()  # Create a file with different name, for sanity check.
    # TODO Tighten test when introducing PyInstaller.exceptions
    with pytest.raises(SystemExit):
        utils.format_binaries_and_datas(datas, str(tmp_path))


def test_format_binaries_and_datas_empty_src(tmp_path):
    # `format_binaries_and_datas()` must disallow empty src in `binaries`/`datas` tuples, as those result in implicit
    # collection of the whole current working directory .
    datas = [('', '.')]
    with pytest.raises(SystemExit, match="Empty SRC is not allowed"):
        utils.format_binaries_and_datas(datas, str(tmp_path))


def test_format_binaries_and_datas_basic(tmp_path):
    # (src, dest) tuples to be passed to format_binaries_and_datas()
    DATAS = (
        ('existing.txt', '.'),
        ('other.txt', 'foo'),
        ('*.log', 'logs'),
        ('a/*.log', 'lll'),
        ('a/here.tex', '.'),
        ('b/[abc].tex', 'tex'),
    )

    # Expected entries; they are listed as (src, dest) tuples for readability; the subsequent code transforms them into
    # (dest, src) tuples format used by format_binaries_and_datas().
    EXPECTED = (
        ('existing.txt', 'existing.txt'),
        ('other.txt', 'foo/other.txt'),
        ('aaa.log', 'logs/aaa.log'),
        ('bbb.log', 'logs/bbb.log'),
        ('a/xxx.log', 'lll/xxx.log'),
        ('a/yyy.log', 'lll/yyy.log'),
        ('a/here.tex', 'here.tex'),
        ('b/a.tex', 'tex/a.tex'),
        ('b/b.tex', 'tex/b.tex'),
    )

    # Normalize separator in source paths
    datas = [(os.path.normpath(src), dest) for src, dest in DATAS]

    # Convert the (src, dest) entries from EXPECTED into (dest, src) format, and turn `src` into full path.
    expected = set()
    for src, dest in EXPECTED:
        src_path = tmp_path / src
        dest_path = pathlib.PurePath(dest)  # Normalize separators.
        # Create the file
        src_path.parent.mkdir(parents=True, exist_ok=True)
        src_path.touch()
        # Expected entry
        expected.add((str(dest_path), str(src_path)))

    # Create some additional files that should not be included.
    (tmp_path / 'not.txt').touch()
    (tmp_path / 'a' / 'not.txt').touch()
    (tmp_path / 'b' / 'not.txt').touch()

    res = utils.format_binaries_and_datas(datas, str(tmp_path))
    assert res == expected


def test_format_binaries_and_datas_with_bracket(tmp_path):
    # See issue #2314: the filename contains brackets which are interpreted by glob().
    DATAS = (
        (('b/[abc].tex'), 'tex'),
    )  # yapf: disable

    EXPECTED = (
        ('b/[abc].tex', 'tex/[abc].tex'),
    )  # yapf: disable

    # Normalize separator in source paths
    datas = [(os.path.normpath(src), dest) for src, dest in DATAS]

    # Convert the (src, dest) entries from EXPECTED into (dest, src) format, and turn `src` into full path.
    expected = set()
    for src, dest in EXPECTED:
        src_path = tmp_path / src
        dest_path = pathlib.PurePath(dest)  # Normalize separators.
        # Create the file
        src_path.parent.mkdir(parents=True, exist_ok=True)
        src_path.touch()
        # Expected entry
        expected.add((str(dest_path), str(src_path)))

    # Create some additional files that should not be included.
    (tmp_path / 'tex').mkdir(parents=True, exist_ok=True)
    (tmp_path / 'tex' / 'not.txt').touch()

    res = utils.format_binaries_and_datas(datas, str(tmp_path))
    assert res == expected


def test_should_include_system_binary():
    python_dir = f'python{sys.version_info.major}.{sys.version_info.minor}'
    CASES = [
        (f'{python_dir}/lib-dynload/any', f'/usr/lib64/{python_dir}/lib-dynload/any', [], True),
        ('libany', '/lib64/libpython.so', [], True),
        ('any', '/lib/python/site-packages/any', [], True),
        ('libany', '/etc/libany', [], True),
        ('libany', '/usr/lib/libany', ['*any*'], True),
        ('libany2', '/lib/libany2', ['libnone*', 'libany*'], True),
        ('libnomatch', '/lib/libnomatch', ['libnone*', 'libany*'], False),
    ]

    for case in CASES:
        tuple = (case[0], case[1])
        excepts = case[2]
        expected = case[3]

        assert utils._should_include_system_binary(tuple, excepts) == expected



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Test for unbuffered stdio (stdout/stderr) mode.
"""

import os
import asyncio

import pytest

from PyInstaller.compat import is_win


@pytest.mark.parametrize('stream_mode', ['binary', 'text'])
@pytest.mark.parametrize('output_stream', ['stdout', 'stderr'])
def test_unbuffered_stdio(tmp_path, output_stream, stream_mode, pyi_builder_spec):
    # Freeze the test program; test_spec() builds the app and runs it, so explicitly set the number of
    # stars to 0 for this run.
    pyi_builder_spec.test_spec('pyi_unbuffered_output.spec', app_args=['--num-stars', '0'])

    # Path to the frozen executable
    executable = os.path.join(tmp_path, 'dist', 'pyi_unbuffered_output', 'pyi_unbuffered_output')

    # Expected number of stars
    EXPECTED_STARS = 5

    # Run the test program via asyncio.SubprocessProtocol and monitor the output.
    class SubprocessDotCounter(asyncio.SubprocessProtocol):
        def __init__(self, loop, output='stdout'):
            self.count = 0
            self.loop = loop
            # Select stdout vs stderr
            assert output in {'stdout', 'stderr'}
            self.out_fd = 1 if output == 'stdout' else 2

        def pipe_data_received(self, fd, data):
            if fd == self.out_fd:
                # Treat any data batch that does not end with the * as irregularity
                if not data.endswith(b'*'):
                    return
                self.count += data.count(b'*')

        def connection_lost(self, exc):
            self.loop.stop()  # end loop.run_forever()

    # Create event loop
    if is_win:
        loop = asyncio.ProactorEventLoop()  # for subprocess' pipes on Windows
    else:
        loop = asyncio.SelectorEventLoop()
    asyncio.set_event_loop(loop)

    counter_proto = SubprocessDotCounter(loop, output=output_stream)

    # Run
    try:
        proc = loop.subprocess_exec(
            lambda: counter_proto,
            executable,
            "--num-stars", str(EXPECTED_STARS),
            "--output-stream", output_stream,
            "--stream-mode", stream_mode
        )  # yapf: disable
        transport, _ = loop.run_until_complete(proc)
        loop.run_forever()
    finally:
        loop.close()
        transport.close()

    # Check the number of received stars
    assert counter_proto.count == EXPECTED_STARS



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import waflib.Tools.asm
from waflib.Tools import ar


def configure(conf):
    conf.find_program(['gas', 'gcc'], var='AS')
    conf.env.AS_TGT_F = ['-c', '-o']
    conf.env.ASLNK_TGT_F = ['-o']
    conf.find_ar()
    conf.load('asm')
    conf.env.ASM_NAME = 'gas'



#-----------------------------------------------------------------------------
# Copyright (c) 2025, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Prevent conditional import of `distutils` in `_osx_support.compiler_fixup()` in python < 3.10 from pulling in
# `distutils`; this function is called only from `distutils` itself, which ensures that the module is available as
# needed. Blocking this import prevents `distutils` (and nowadays `setuptools`) from being pulled into even very
# basic applications when built with python < 3.10.
#
# See: https://github.com/python/cpython/blob/f3994ade31a563d49806cf6a681d1b1115fccaa3/Lib/_osx_support.py#L430-L434

excludedimports = ['distutils']



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re, traceback, sys
from waflib import Utils, ansiterm
if not os.environ.get('NOSYNC', False):
    if sys.stdout.isatty() and id(sys.stdout) == id(sys.__stdout__):
        sys.stdout = ansiterm.AnsiTerm(sys.stdout)
    if sys.stderr.isatty() and id(sys.stderr) == id(sys.__stderr__):
        sys.stderr = ansiterm.AnsiTerm(sys.stderr)
import logging

LOG_FORMAT = os.environ.get('WAF_LOG_FORMAT', '%(asctime)s %(c1)s%(zone)s%(c2)s %(message)s')
HOUR_FORMAT = os.environ.get('WAF_HOUR_FORMAT', '%H:%M:%S')
zones = []
verbose = 0
colors_lst = {
    'USE': True,
    'BOLD': '\x1b[01;1m',
    'RED': '\x1b[01;31m',
    'GREEN': '\x1b[32m',
    'YELLOW': '\x1b[33m',
    'PINK': '\x1b[35m',
    'BLUE': '\x1b[01;34m',
    'CYAN': '\x1b[36m',
    'GREY': '\x1b[37m',
    'NORMAL': '\x1b[0m',
    'cursor_on': '\x1b[?25h',
    'cursor_off': '\x1b[?25l',
}
indicator = '\r\x1b[K%s%s%s'
try:
    unicode
except NameError:
    unicode = None


def enable_colors(use):
    if use == 1:
        if not (sys.stderr.isatty() or sys.stdout.isatty()):
            use = 0
        if Utils.is_win32 and os.name != 'java':
            term = os.environ.get('TERM', '')
        else:
            term = os.environ.get('TERM', 'dumb')
        if term in ('dumb', 'emacs'):
            use = 0
    if use >= 1:
        os.environ['TERM'] = 'vt100'
    colors_lst['USE'] = use


try:
    get_term_cols = ansiterm.get_term_cols
except AttributeError:

    def get_term_cols():
        return 80


get_term_cols.__doc__ = """
	Returns the console width in characters.

	:return: the number of characters per line
	:rtype: int
	"""


def get_color(cl):
    if colors_lst['USE']:
        return colors_lst.get(cl, '')
    return ''


class color_dict(object):
    def __getattr__(self, a):
        return get_color(a)

    def __call__(self, a):
        return get_color(a)


colors = color_dict()
re_log = re.compile(r'(\w+): (.*)', re.M)


class log_filter(logging.Filter):
    def __init__(self, name=''):
        logging.Filter.__init__(self, name)

    def filter(self, rec):
        rec.zone = rec.module
        if rec.levelno >= logging.INFO:
            return True
        m = re_log.match(rec.msg)
        if m:
            rec.zone = m.group(1)
            rec.msg = m.group(2)
        if zones:
            return getattr(rec, 'zone', '') in zones or '*' in zones
        elif not verbose > 2:
            return False
        return True


class log_handler(logging.StreamHandler):
    def emit(self, record):
        try:
            try:
                self.stream = record.stream
            except AttributeError:
                if record.levelno >= logging.WARNING:
                    record.stream = self.stream = sys.stderr
                else:
                    record.stream = self.stream = sys.stdout
            self.emit_override(record)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def emit_override(self, record, **kw):
        self.terminator = getattr(record, 'terminator', '\n')
        stream = self.stream
        if unicode:
            msg = self.formatter.format(record)
            fs = '%s' + self.terminator
            try:
                if (isinstance(msg, unicode) and getattr(stream, 'encoding', None)):
                    fs = fs.decode(stream.encoding)
                    try:
                        stream.write(fs % msg)
                    except UnicodeEncodeError:
                        stream.write((fs % msg).encode(stream.encoding))
                else:
                    stream.write(fs % msg)
            except UnicodeError:
                stream.write((fs % msg).encode('utf-8'))
        else:
            logging.StreamHandler.emit(self, record)


class formatter(logging.Formatter):
    def __init__(self):
        logging.Formatter.__init__(self, LOG_FORMAT, HOUR_FORMAT)

    def format(self, rec):
        try:
            msg = rec.msg.decode('utf-8')
        except Exception:
            msg = rec.msg
        use = colors_lst['USE']
        if (use == 1 and rec.stream.isatty()) or use == 2:
            c1 = getattr(rec, 'c1', None)
            if c1 is None:
                c1 = ''
                if rec.levelno >= logging.ERROR:
                    c1 = colors.RED
                elif rec.levelno >= logging.WARNING:
                    c1 = colors.YELLOW
                elif rec.levelno >= logging.INFO:
                    c1 = colors.GREEN
            c2 = getattr(rec, 'c2', colors.NORMAL)
            msg = '%s%s%s' % (c1, msg, c2)
        else:
            msg = re.sub(r'\r(?!\n)|\x1B\[(K|.*?(m|h|l))', '', msg)
        if rec.levelno >= logging.INFO:
            if rec.args:
                try:
                    return msg % rec.args
                except UnicodeDecodeError:
                    return msg.encode('utf-8') % rec.args
            return msg
        rec.msg = msg
        rec.c1 = colors.PINK
        rec.c2 = colors.NORMAL
        return logging.Formatter.format(self, rec)


log = None


def debug(*k, **kw):
    if verbose:
        k = list(k)
        k[0] = k[0].replace('\n', ' ')
        log.debug(*k, **kw)


def error(*k, **kw):
    log.error(*k, **kw)
    if verbose > 2:
        st = traceback.extract_stack()
        if st:
            st = st[:-1]
            buf = []
            for filename, lineno, name, line in st:
                buf.append('  File %r, line %d, in %s' % (filename, lineno, name))
                if line:
                    buf.append('	%s' % line.strip())
            if buf:
                log.error('\n'.join(buf))


def warn(*k, **kw):
    log.warning(*k, **kw)


def info(*k, **kw):
    log.info(*k, **kw)


def init_log():
    global log
    log = logging.getLogger('waflib')
    log.handlers = []
    log.filters = []
    hdlr = log_handler()
    hdlr.setFormatter(formatter())
    log.addHandler(hdlr)
    log.addFilter(log_filter())
    log.setLevel(logging.DEBUG)


def make_logger(path, name):
    logger = logging.getLogger(name)
    if sys.hexversion > 0x3000000:
        encoding = sys.stdout.encoding
    else:
        encoding = None
    hdlr = logging.FileHandler(path, 'w', encoding=encoding)
    formatter = logging.Formatter('%(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.DEBUG)
    return logger


def make_mem_logger(name, to_log, size=8192):
    from logging.handlers import MemoryHandler
    logger = logging.getLogger(name)
    hdlr = MemoryHandler(size, target=to_log)
    formatter = logging.Formatter('%(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.memhandler = hdlr
    logger.setLevel(logging.DEBUG)
    return logger


def free_logger(logger):
    try:
        for x in logger.handlers:
            x.close()
            logger.removeHandler(x)
    except Exception:
        pass


def pprint(col, msg, label='', sep='\n'):
    info('%s%s%s %s', colors(col), msg, colors.NORMAL, label, extra={'terminator': sep})



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import argparse

import pytest

from PyInstaller.building import makespec


def test_make_variable_path():
    p = os.path.join(makespec.HOMEPATH, "aaa", "bbb", "ccc")
    assert (makespec.make_variable_path(p) == ("HOMEPATH", os.path.join("aaa", "bbb", "ccc")))


def test_make_variable_path_regression():
    p = os.path.join(makespec.HOMEPATH + "aaa", "bbb", "ccc")
    assert makespec.make_variable_path(p) == (None, p)


def test_Path_constructor():
    p = makespec.Path("aaa", "bbb", "ccc")
    assert p.path == os.path.join("aaa", "bbb", "ccc")


def test_Path_repr():
    p = makespec.Path(makespec.HOMEPATH, "aaa", "bbb", "ccc")
    assert p.path == os.path.join(makespec.HOMEPATH, "aaa", "bbb", "ccc")
    assert (repr(p) == "os.path.join(HOMEPATH,%r)" % os.path.join("aaa", "bbb", "ccc"))


def test_Path_repr_relative():
    p = makespec.Path("aaa", "bbb", "ccc.py")
    assert p.path == os.path.join("aaa", "bbb", "ccc.py")
    assert repr(p) == "%r" % os.path.join("aaa", "bbb", "ccc.py")


def test_Path_regression():
    p = makespec.Path(makespec.HOMEPATH + "-aaa", "bbb", "ccc")
    assert p.path == os.path.join(makespec.HOMEPATH + "-aaa", "bbb", "ccc")
    assert (repr(p) == repr(os.path.join(makespec.HOMEPATH + "-aaa", "bbb", "ccc")))


def test_add_data(capsys):
    """
    Test CLI parsing of --add-data and --add-binary.
    """
    parser = argparse.ArgumentParser()
    makespec.__add_options(parser)

    assert parser.parse_args([]).datas == []
    assert parser.parse_args(["--add-data", "/foo/bar:."]).datas == [("/foo/bar", ".")]
    assert parser.parse_args([r"--add-data=C:\foo\bar:baz"]).datas == [(r"C:\foo\bar", "baz")]
    assert parser.parse_args([r"--add-data=c:/foo/bar:baz"]).datas == [(r"c:/foo/bar", "baz")]
    assert parser.parse_args([r"--add-data=/foo/:bar"]).datas == [("/foo/", "bar")]

    for args in [["--add-data", "foo/bar"], ["--add-data", "C:/foo/bar"]]:
        with pytest.raises(SystemExit):
            parser.parse_args(args)
        assert '--add-data: Wrong syntax, should be --add-data=SOURCE:DEST' in capsys.readouterr().err

    if os.pathsep == ";":
        assert parser.parse_args(["--add-data", "foo;."]).datas == [("foo", ".")]
    else:
        assert parser.parse_args(["--add-data", "foo;bar:."]).datas == [("foo;bar", ".")]

    with pytest.raises(SystemExit):
        parser.parse_args(["--add-data", "foo:"])
    assert '--add-data: You have to specify both SOURCE and DEST' in capsys.readouterr().err

    options = parser.parse_args(["--add-data=a:b", "--add-data=c:d", "--add-binary=e:f"])
    assert options.datas == [("a", "b"), ("c", "d")]
    assert options.binaries == [("e", "f")]



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import subprocess

import pytest

from PyInstaller.utils.tests import importorskip


def test_ascii_path(pyi_builder):
    dist_dir = str(pyi_builder._dist_dir)

    dist_dir_ascii = dist_dir.encode('ascii', 'replace').decode('ascii')
    if dist_dir != dist_dir_ascii:
        pytest.skip(reason="Default build path not ASCII, skipping...")

    pyi_builder.test_script('pyi_path_encoding.py')


@pytest.mark.linux
def test_linux_non_unicode_path(pyi_builder, monkeypatch):
    # If we set the locale to 'C', mbstowcs should be completely useless. This test verifies that _Py_char2wchar will
    # decode the "undecodable" bytes and will decode even filenames that weren't encoded with the locale encoding.
    unicode_filename = 'ěščřžýáíé日本語'

    pyi_builder._dist_dir = pyi_builder._dist_dir / unicode_filename

    tmp = pyi_builder._tmp_path / f"{unicode_filename}_TMP"
    monkeypatch.setenv('LC_ALL', 'C')
    monkeypatch.setenv('TMPDIR', str(tmp))
    monkeypatch.setenv('TMP', str(tmp))

    pyi_builder.test_script('pyi_path_encoding.py')


@pytest.mark.darwin
@pytest.mark.linux
def test_osx_linux_unicode_path(pyi_builder, monkeypatch):
    # Mac and Linux should handle 'unicode' type filenames without problem.
    unicode_filename = 'ěščřžýáíé日本語'

    pyi_builder._dist_dir = pyi_builder._dist_dir / unicode_filename

    tmp = pyi_builder._tmp_path / f"{unicode_filename}_TMP"
    monkeypatch.setenv('TMPDIR', str(tmp))
    monkeypatch.setenv('TMP', str(tmp))

    pyi_builder.test_script('pyi_path_encoding.py')


@pytest.mark.win32
def test_win_codepage_path(pyi_builder, monkeypatch):
    # Create some bytes and decode with the current codepage to get a filename that is guaranteed to encode with the
    # current codepage. Assumes a one-byte codepage, i.e., not cp937 (shift-JIS) which is multibyte.
    cp_filename = bytes(bytearray(range(0x80, 0x86))).decode('mbcs')

    pyi_builder._dist_dir = pyi_builder._dist_dir / cp_filename

    tmp = pyi_builder._tmp_path / f"{cp_filename}_TMP"
    monkeypatch.setenv('TMPDIR', str(tmp))
    monkeypatch.setenv('TMP', str(tmp))

    pyi_builder.test_script('pyi_path_encoding.py')


@pytest.mark.win32
def test_win_codepage_path_disabled_shortfilename(pyi_builder, monkeypatch):
    # Create some bytes and decode with the current codepage to get a filename that is guaranteed to encode with the
    # current codepage. Assumes a one-byte codepage, i.e., not cp937 (shift-JIS) which is multibyte.
    cp_filename = bytes(bytearray(range(0x80, 0x86))).decode('mbcs')

    pyi_builder._dist_dir = pyi_builder._dist_dir / cp_filename

    # Try to remove ShortFileName from this folder using `fsutil`. Requires admin privileges, so `xfail` if we do not
    # have them. `8dot3name strip` only affects subfolders, so pass the folder containing our codepage filename.
    pyi_builder._dist_dir.mkdir(parents=True)  # Ensure directory exists so we can modify it with `fsutil`.
    if subprocess.call(['fsutil', '8dot3name', 'strip', str(pyi_builder._dist_dir)]):
        pytest.xfail("Administrator privileges required to strip ShortFileName.")

    tmp = pyi_builder._tmp_path / f"{cp_filename}_TMP"
    monkeypatch.setenv('TMPDIR', str(tmp))
    monkeypatch.setenv('TMP', str(tmp))

    pyi_builder.test_script('pyi_path_encoding.py')


@pytest.mark.win32
def test_win_non_codepage_path(pyi_builder, monkeypatch):
    # Both eastern European and Japanese characters - no codepage should encode this.
    non_cp_filename = 'ěščřžýáíé日本語'

    # Codepage encoding would replace some of these chars with "???".
    pyi_builder._dist_dir = pyi_builder._dist_dir / non_cp_filename

    # To test what happens with a non-ANSI tempdir, we will also need to pass the TMP environ as wide chars.
    tmp = pyi_builder._tmp_path / f"{non_cp_filename}_TMP"
    monkeypatch.setenv('TMPDIR', str(tmp))
    monkeypatch.setenv('TMP', str(tmp))

    pyi_builder.test_script('pyi_path_encoding.py')


@pytest.mark.win32
@importorskip('win32api')
def test_win_py3_no_shortpathname(pyi_builder):
    pyi_builder.test_script('pyi_win_py3_no_shortpathname.py')


@pytest.mark.win32
@importorskip('win32api')
def test_win_TEMP_has_shortpathname(pyi_builder, monkeypatch, tmp_path):
    """
    Test if script if pass if $TMP holds a short path name.
    """
    tmp = tmp_path / "longlongfilename" / "xxx"
    tmp.mkdir(parents=True, exist_ok=True)
    import win32api
    tmp = win32api.GetShortPathName(str(tmp))
    monkeypatch.setenv("TMP", tmp)
    monkeypatch.setenv("TEMP", tmp)
    pyi_builder.test_script('pyi_win_py3_no_shortpathname.py')



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import sys
import multiprocessing


def test_function(semaphore):
    semaphore.acquire()
    print('In subprocess')
    semaphore.release()


def main(start_method):
    multiprocessing.set_start_method(start_method)

    print('In main')
    sys.stdout.flush()
    semaphore = multiprocessing.Semaphore()
    semaphore.acquire()
    process = multiprocessing.Process(target=test_function, args=[semaphore])
    process.start()
    semaphore.release()
    process.join()


if __name__ == '__main__':
    multiprocessing.freeze_support()

    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <start-method>")
        sys.exit(1)

    main(sys.argv[1])



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Utils, Task, Errors
from waflib.Tools import ccroot, fc_config, fc_scan
from waflib.TaskGen import extension
from waflib.Configure import conf

ccroot.USELIB_VARS['fc'] = set(['FCFLAGS', 'DEFINES', 'INCLUDES', 'FCPPFLAGS'])
ccroot.USELIB_VARS['fcprogram_test'] = ccroot.USELIB_VARS['fcprogram'] = set([
    'LIB', 'STLIB', 'LIBPATH', 'STLIBPATH', 'LINKFLAGS', 'RPATH', 'LINKDEPS', 'LDFLAGS'
])
ccroot.USELIB_VARS['fcshlib'] = set([
    'LIB', 'STLIB', 'LIBPATH', 'STLIBPATH', 'LINKFLAGS', 'RPATH', 'LINKDEPS', 'LDFLAGS'
])
ccroot.USELIB_VARS['fcstlib'] = set(['ARFLAGS', 'LINKDEPS'])


@extension('.f', '.F', '.f90', '.F90', '.for', '.FOR', '.f95', '.F95', '.f03', '.F03', '.f08', '.F08')
def fc_hook(self, node):
    return self.create_compiled_task('fc', node)


@conf
def modfile(conf, name):
    if name.find(':') >= 0:
        separator = conf.env.FC_SUBMOD_SEPARATOR or '@'
        modpath = name.split(':')
        modname = modpath[0] + separator + modpath[-1]
        suffix = conf.env.FC_SUBMOD_SUFFIX or '.smod'
    else:
        modname = name
        suffix = '.mod'
    return {
        'lower': modname.lower() + suffix.lower(),
        'lower.MOD': modname.lower() + suffix.upper(),
        'UPPER.mod': modname.upper() + suffix.lower(),
        'UPPER': modname.upper() + suffix.upper()
    }[conf.env.FC_MOD_CAPITALIZATION or 'lower']


def get_fortran_tasks(tsk):
    bld = tsk.generator.bld
    tasks = bld.get_tasks_group(bld.get_group_idx(tsk.generator))
    return [
        x for x in tasks
        if isinstance(x, fc) and not getattr(x, 'nomod', None) and not getattr(x, 'mod_fortran_done', None)
    ]


class fc(Task.Task):
    color = 'GREEN'
    run_str = '${FC} ${FCFLAGS} ${FCINCPATH_ST:INCPATHS} ${FCDEFINES_ST:DEFINES} ${_FCMODOUTFLAGS} ${FC_TGT_F}${TGT[0].abspath()} ${FC_SRC_F}${SRC[0].abspath()} ${FCPPFLAGS}'
    vars = ["FORTRANMODPATHFLAG"]

    def scan(self):
        tmp = fc_scan.fortran_parser(self.generator.includes_nodes)
        tmp.task = self
        tmp.start(self.inputs[0])
        return (tmp.nodes, tmp.names)

    def runnable_status(self):
        if getattr(self, 'mod_fortran_done', None):
            return super(fc, self).runnable_status()
        bld = self.generator.bld
        lst = get_fortran_tasks(self)
        for tsk in lst:
            tsk.mod_fortran_done = True
        for tsk in lst:
            ret = tsk.runnable_status()
            if ret == Task.ASK_LATER:
                for x in lst:
                    x.mod_fortran_done = None
                return Task.ASK_LATER
        ins = Utils.defaultdict(set)
        outs = Utils.defaultdict(set)
        for tsk in lst:
            key = tsk.uid()
            for x in bld.raw_deps[key]:
                if x.startswith('MOD@'):
                    name = bld.modfile(x.replace('MOD@', ''))
                    node = bld.srcnode.find_or_declare(name)
                    tsk.set_outputs(node)
                    outs[node].add(tsk)
        for tsk in lst:
            key = tsk.uid()
            for x in bld.raw_deps[key]:
                if x.startswith('USE@'):
                    name = bld.modfile(x.replace('USE@', ''))
                    node = bld.srcnode.find_resource(name)
                    if node and node not in tsk.outputs:
                        if not node in bld.node_deps[key]:
                            bld.node_deps[key].append(node)
                        ins[node].add(tsk)
        for k in ins.keys():
            for a in ins[k]:
                a.run_after.update(outs[k])
                for x in outs[k]:
                    self.generator.bld.producer.revdeps[x].add(a)
                tmp = []
                for t in outs[k]:
                    tmp.extend(t.outputs)
                a.dep_nodes.extend(tmp)
                a.dep_nodes.sort(key=lambda x: x.abspath())
        for tsk in lst:
            try:
                delattr(tsk, 'cache_sig')
            except AttributeError:
                pass
        return super(fc, self).runnable_status()


class fcprogram(ccroot.link_task):
    color = 'YELLOW'
    run_str = '${FC} ${LINKFLAGS} ${FCLNK_SRC_F}${SRC} ${FCLNK_TGT_F}${TGT[0].abspath()} ${RPATH_ST:RPATH} ${FCSTLIB_MARKER} ${FCSTLIBPATH_ST:STLIBPATH} ${FCSTLIB_ST:STLIB} ${FCSHLIB_MARKER} ${FCLIBPATH_ST:LIBPATH} ${FCLIB_ST:LIB} ${LDFLAGS}'
    inst_to = '${BINDIR}'


class fcshlib(fcprogram):
    inst_to = '${LIBDIR}'


class fcstlib(ccroot.stlink_task):
    pass


class fcprogram_test(fcprogram):
    def runnable_status(self):
        ret = super(fcprogram_test, self).runnable_status()
        if ret == Task.SKIP_ME:
            ret = Task.RUN_ME
        return ret

    def exec_command(self, cmd, **kw):
        bld = self.generator.bld
        kw['shell'] = isinstance(cmd, str)
        kw['stdout'] = kw['stderr'] = Utils.subprocess.PIPE
        kw['cwd'] = self.get_cwd()
        bld.out = bld.err = ''
        bld.to_log('command: %s\n' % cmd)
        kw['output'] = 0
        try:
            (bld.out, bld.err) = bld.cmd_and_log(cmd, **kw)
        except Errors.WafError:
            return -1
        if bld.out:
            bld.to_log('out: %s\n' % bld.out)
        if bld.err:
            bld.to_log('err: %s\n' % bld.err)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import sys
import multiprocessing


def f(x):
    return x * x


def main(start_method):
    multiprocessing.set_start_method(start_method)

    # Start a pool with 4 worker processes.
    with multiprocessing.Pool(processes=4) as pool:
        print('Evaluate f(10) asynchronously...')
        result = pool.apply_async(f, [10])
        assert result.get() == 100

        print('Evaluate f(0..9)...')
        assert pool.map(f, range(10)) == [x**2 for x in range(10)]


if __name__ == '__main__':
    multiprocessing.freeze_support()

    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <start-method>")
        sys.exit(1)

    main(sys.argv[1])



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib import Utils


def filter_comments(filename):
    txt = Utils.readf(filename)
    i = 0
    buf = []
    max = len(txt)
    begin = 0
    while i < max:
        c = txt[i]
        if c == '"' or c == "'":
            buf.append(txt[begin:i])
            delim = c
            i += 1
            while i < max:
                c = txt[i]
                if c == delim:
                    break
                elif c == '\\':
                    i += 1
                i += 1
            i += 1
            begin = i
        elif c == '/':
            buf.append(txt[begin:i])
            i += 1
            if i == max:
                break
            c = txt[i]
            if c == '+':
                i += 1
                nesting = 1
                c = None
                while i < max:
                    prev = c
                    c = txt[i]
                    if prev == '/' and c == '+':
                        nesting += 1
                        c = None
                    elif prev == '+' and c == '/':
                        nesting -= 1
                        if nesting == 0:
                            break
                        c = None
                    i += 1
            elif c == '*':
                i += 1
                c = None
                while i < max:
                    prev = c
                    c = txt[i]
                    if prev == '*' and c == '/':
                        break
                    i += 1
            elif c == '/':
                i += 1
                while i < max and txt[i] != '\n':
                    i += 1
            else:
                begin = i - 1
                continue
            i += 1
            begin = i
            buf.append(' ')
        else:
            i += 1
    buf.append(txt[begin:])
    return buf


class d_parser(object):
    def __init__(self, env, incpaths):
        self.allnames = []
        self.re_module = re.compile(r"module\s+([^;]+)")
        self.re_import = re.compile(r"import\s+([^;]+)")
        self.re_import_bindings = re.compile("([^:]+):(.*)")
        self.re_import_alias = re.compile("[^=]+=(.+)")
        self.env = env
        self.nodes = []
        self.names = []
        self.incpaths = incpaths

    def tryfind(self, filename):
        found = 0
        for n in self.incpaths:
            found = n.find_resource(filename.replace('.', '/') + '.d')
            if found:
                self.nodes.append(found)
                self.waiting.append(found)
                break
        if not found:
            if not filename in self.names:
                self.names.append(filename)

    def get_strings(self, code):
        self.module = ''
        lst = []
        mod_name = self.re_module.search(code)
        if mod_name:
            self.module = re.sub(r'\s+', '', mod_name.group(1))
        import_iterator = self.re_import.finditer(code)
        if import_iterator:
            for import_match in import_iterator:
                import_match_str = re.sub(r'\s+', '', import_match.group(1))
                bindings_match = self.re_import_bindings.match(import_match_str)
                if bindings_match:
                    import_match_str = bindings_match.group(1)
                matches = import_match_str.split(',')
                for match in matches:
                    alias_match = self.re_import_alias.match(match)
                    if alias_match:
                        match = alias_match.group(1)
                    lst.append(match)
        return lst

    def start(self, node):
        self.waiting = [node]
        while self.waiting:
            nd = self.waiting.pop(0)
            self.iter(nd)

    def iter(self, node):
        path = node.abspath()
        code = "".join(filter_comments(path))
        names = self.get_strings(code)
        for x in names:
            if x in self.allnames:
                continue
            self.allnames.append(x)
            self.tryfind(x)


def scan(self):
    env = self.env
    gruik = d_parser(env, self.generator.includes_nodes)
    node = self.inputs[0]
    gruik.start(node)
    nodes = gruik.nodes
    names = gruik.names
    return (nodes, names)



#-----------------------------------------------------------------------------
# Copyright (c) 2019-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
`importlib_resources` is a backport of the 3.9+ module `importlib.resources`
"""

from PyInstaller.utils.hooks import check_requirement, collect_data_files

# Prior to v1.2.0, a `version.txt` file is used to set __version__. Later versions use `importlib.metadata`.
if check_requirement("importlib_resources < 1.2.0"):
    datas = collect_data_files("importlib_resources", includes=["version.txt"])

if check_requirement("importlib_resources >= 1.3.1"):
    hiddenimports = ['importlib_resources.trees']



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

name = 'pyi_testmod_relimp.relimp1'

from . import relimp2 as upper  # noqa: E402
from .pyi_testmod_relimp import relimp2 as lower  # noqa: E402

assert upper.name == 'pyi_testmod_relimp.relimp2'
assert lower.name == 'pyi_testmod_relimp.pyi_testmod_relimp.relimp2'

if upper.__name__ == lower.__name__:
    raise SystemExit("Imported the same module")

if upper.__file__ == lower.__file__:
    raise SystemExit("Imported the same file")



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Decorators for skipping PyInstaller tests when specific requirements are not met.
"""

import inspect
import sys
import textwrap

import pytest

from PyInstaller.utils.hooks import check_requirement

# Wrap some pytest decorators to be consistent in tests.
parametrize = pytest.mark.parametrize
skipif = pytest.mark.skipif
xfail = pytest.mark.xfail
skip = pytest.mark.skip

# Use these decorators to use the `pyi_builder` fixture only in onedir or only in onefile mode instead of both.
onedir_only = pytest.mark.parametrize('pyi_builder', ['onedir'], indirect=True)
onefile_only = pytest.mark.parametrize('pyi_builder', ['onefile'], indirect=True)


def importorskip(package: str):
    """
    Skip a decorated test if **package** is not importable.

    Arguments:
        package:
            The name of the module. May be anything that is allowed after the ``import`` keyword. e.g. 'numpy' or
            'PIL.Image'.
    Returns:
        A pytest marker which either skips the test or does nothing.

    This function intentionally does not import the module. Doing so can lead to `sys.path` and `PATH` being
    polluted, which then breaks later builds.
    """
    if not importable(package):
        return pytest.mark.skip(f"Can't import '{package}'.")
    return pytest.mark.skipif(False, reason=f"Don't skip: '{package}' is importable.")


def importable(package: str):
    from importlib.util import find_spec

    # The find_spec() function is used by the importlib machinery to locate a module to import. Using it finds the
    # module but does not run it. Unfortunately, it does import parent modules to check submodules.
    if "." in package:
        # Using subprocesses is slow. If the top level module doesn't exist then we can skip it.
        if not importable(package.split(".")[0]):
            return False
        # This is a submodule, import it in isolation.
        from subprocess import DEVNULL, run
        return run([sys.executable, "-c", "import " + package], stdout=DEVNULL, stderr=DEVNULL).returncode == 0

    return find_spec(package) is not None


def requires(requirement: str):
    """
    Mark a test to be skipped if **requirement** is not satisfied.

    Args:
        requirement:
            A distribution name and optional version specifier(s). See :func:`PyInstaller.utils.hooks.check_requirement`
            which this argument is forwarded to.
    Returns:
        Either a skip marker or a dummy marker.

    This function operates on distribution metadata, and does not import any modules.
    """
    if check_requirement(requirement):
        return pytest.mark.skipif(False, reason=f"Don't skip: '{requirement}' is satisfied.")
    else:
        return pytest.mark.skip(f"Requires {requirement}.")


def gen_sourcefile(tmp_path, source, test_id=None):
    """
    Generate a source file for testing.

    The source will be written into a file named like the test-function. This file will then be passed to
    `test_script`. If you need other related file, e.g. as `.toc`-file for testing the content, put it at at the
    normal place. Just mind to take the basnename from the test-function's name.

    :param script: Source code to create executable from. This will be saved into a temporary file which is then
                   passed on to `test_script`.

    :param test_id: Test-id for parametrized tests. If given, it will be appended to the script filename,
                    separated by two underscores.
    """
    testname = inspect.stack()[1][3]
    if test_id:
        # For parametrized test append the test-id.
        testname = testname + '__' + test_id

    # Periods are not allowed in Python module names.
    testname = testname.replace('.', '_')
    scriptfile = tmp_path / (testname + '.py')
    source = textwrap.dedent(source)
    scriptfile.write_text(source, encoding='utf-8')
    return scriptfile



"""
pyi_dummy_module

This module exists only so it can be imported and its __file__ inspected in the test `test_compiled_filenames`.
"""


def dummy():
    pass


class DummyClass(object):
    def dummyMethod(self):
        pass



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# A simple program for testing unbuffered mode of python stdout and stderr streams (both binary and text layers).
#
# The program periodically prints star ('*') character to a single line on the specified output stream (stdout or
# stderr). Once the selected number of stars have been printed, the end-of-transmission is signalled by 'E' character.
#
# In the unbuffered mode, the caller should receive the characters individually, and have enough time to process them.
# In the buffered mode, all printed characters including the terminating E will be received at once.

import argparse
import time
import sys

# Argument parser
parser = argparse.ArgumentParser(description="Unbuffered stdio test")
parser.add_argument(
    '--num-stars',
    type=int,
    default=5,
    help="Number of star characters to print.",
)
parser.add_argument(
    '--output-stream',
    type=str,
    default='stdout',
    help="Output stream ('stdout' or 'stderr')",
)
parser.add_argument(
    '--stream-mode',
    type=str,
    default='text',
    help="Output stream mode ('text' or 'binary')",
)
args = parser.parse_args()

# Select output stream and mode
assert args.output_stream in {'stdout', 'stderr'}, f"Invalid output stream: {args.output_stream}!"
assert args.stream_mode in {'text', 'binary'}, f"Invalid output stream mode: {args.stream_mode}!"

stream = sys.stdout if args.output_stream == 'stdout' else sys.stderr
if args.stream_mode == 'binary':
    stream = stream.buffer  # Use binary layer
    STAR = b'*'
    EOT = b'E'
else:
    STAR = '*'
    EOT = 'E'

# Print the specified number of stars in a single line
for i in range(args.num_stars):
    stream.write(STAR)
    time.sleep(1)

# End-of-transmission
stream.write(EOT)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Task
from waflib.TaskGen import extension


class bison(Task.Task):
    color = 'BLUE'
    run_str = '${BISON} ${BISONFLAGS} ${SRC[0].abspath()} -o ${TGT[0].name}'
    ext_out = ['.h']


@extension('.y', '.yc', '.yy')
def big_bison(self, node):
    has_h = '-d' in self.env.BISONFLAGS
    outs = []
    if node.name.endswith('.yc'):
        outs.append(node.change_ext('.tab.cc'))
        if has_h:
            outs.append(node.change_ext('.tab.hh'))
    else:
        outs.append(node.change_ext('.tab.c'))
        if has_h:
            outs.append(node.change_ext('.tab.h'))
    tsk = self.create_task('bison', node, outs)
    tsk.cwd = node.parent.get_bld()
    self.source.append(outs[0])


def configure(conf):
    conf.find_program('bison', var='BISON')
    conf.env.BISONFLAGS = ['-d']



# Feature #3: attempt to import submodule from mymodule_feature3
feature3_available = False
try:
    import mymodule_feature3.submodule1  # noqa: F401
    feature3_available = True
except ImportError:
    pass



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks import collect_submodules

# The layers to load can be configured using scapy's conf.load_layers.
#  from scapy.config import conf; print(conf.load_layers)
# I decided not to use this, but to include all layer modules. The reason is: When building the package, load_layers may
# not include all the layer modules the program will use later.

hiddenimports = collect_submodules('scapy.layers')



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import sys

# Bootloader should override/set LD_LIBRARY_PATH (Linux, Solaris, AIX, and other Unixes only).
libpath = sys._MEIPASS

# The name of the environment variable used to define the path where the OS should search for dynamic libraries.
if sys.platform.startswith('aix'):
    libpath_var_name = 'LIBPATH'
else:
    libpath_var_name = 'LD_LIBRARY_PATH'

orig_libpath = os.environ.get(libpath_var_name + "_ORIG")
if orig_libpath:
    libpath += ':' + orig_libpath
print(('LD_LIBRARY_PATH expected: ' + libpath))

libpath_from_env = os.environ.get(libpath_var_name)
print(('LD_LIBRARY_PATH  current: ' + libpath_from_env))

if not libpath == libpath_from_env:
    raise SystemExit("Expected LD_LIBRARY_PATH doesn't match.")

# LD_LIBRARY_PATH set by bootloader should be an absolute path.
part1 = libpath_from_env.split(os.pathsep, 1)[0]
print('First path element:', part1)
if not os.path.isabs(part1):
    raise SystemExit("LD_LIBRARY_PATH is not an absolute path.")



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# A test script for validation of importlib.resources / importlib_resources resource reader implementation.
#
# The test package has the following structure:
#
# pyi_pkgres_testpkg/
# ├── a.py
# ├── b.py
# ├── __init__.py
# ├── subpkg1
# │   ├── c.py
# │   ├── data
# │   │   ├── entry1.txt
# │   │   ├── entry2.md
# │   │   ├── entry3.rst
# │   │   └── extra
# │   │       └── extra_entry1.json
# │   ├── d.py
# │   └── __init__.py
# ├── subpkg2
# │   ├── __init__.py
# │   ├── mod.py
# │   └── subsubpkg21
# │       ├── __init__.py
# │       └── mod.py
# └── subpkg3
#     ├── _datafile.json
#     └── __init__.py
#
# When run as unfrozen script, this script can be used to check the behavior of "native" resource readers (or
# compatibility adapters for python-provided loaders) that are provided by importlib.resources (or its
# importlib_resources back-port for python 3.8 and earlier).
#
# When run as a frozen application, this script validates the behavior of the resource reader implemented by
# PyInstaller. Due to transitivity of test results, this script running without errors both as a native script and
# as a frozen application serves as proof of conformance for the PyInstaller's reader.
#
# NOTE: this test is modelled after pkg_resources test, but it does not support zipped eggs. Similarly, it assumes that
# in the frozen version, only access to on-disk resources is provided (either by importlib.resources's adapter, or by
# PyInstaller's custom resource reader, if/when implemented). Without custom resource reader, the frozen version of the
# test seems to work with the importlib.resources (python 3.9 and later) but not with the importlib_resources (python
# 3.8 and earlier), due to implementation differences.
#
# NOTE: functions `contents`, `is_resource`, `path`, `read_binary`, and `read_text` have been deprecated in python 3.11
# stdlib (importlib_resources 5.7) and removed in python 3.13 stdlib (importlib_resources 6.0). We test these based on
# their availability.
#
# NOTE: the afore-mentioned functions were brought back in importlib_resources 6.4.0 (and the change was backported to
# python 3.13b1). Their behavior is changed; if they are given a sub-path as a resource name, they do not raise a
# ValueError.

import sys
import pathlib

# Ensure importlib.resources from python 3.9 stdlib or equivalent importlib_resources >= 1.3.
try:
    import importlib.resources as importlib_resources
    if not hasattr(importlib_resources, "files"):
        raise ImportError("Built-in importlib.resources is too old!")
    is_builtin = True
    package_name = 'importlib.resources'
    print("Using built-in importlib.resources...")
except ImportError:
    import importlib_resources
    is_builtin = False
    package_name = 'importlib_resources'
    print("Using backported importlib_resources...")

is_frozen = getattr(sys, 'frozen', False)

# Try to determine the presence of new functional API, which was introduced in importlib_resources 6.4. As per note at
# the top of the file, the new functional API allows sub-paths to be passed as resource names. In importlib_resources
# 6.4, the module was named `functional`, although it will likely be renamed to `_functional` in subsequent versions
# (see https://github.com/python/importlib_resources/pull/306). The functional API was also backported to python 3.13b1
# stdlib, where the module is already called `_functional`.
new_functional_api = hasattr(importlib_resources, 'functional') or hasattr(importlib_resources, '_functional')

########################################################################
#          Validate behavior of importlib.resources.contents()         #
########################################################################
if hasattr(importlib_resources, "contents"):
    print(f"Testing {package_name}.contents()...")

    def _contents_test(pkgname, expected):
        # For frozen application, remove .py files from expected results.
        if is_frozen:
            expected = [x for x in expected if not x.endswith('.py')]
        # List the content
        content = list(importlib_resources.contents(pkgname))
        # Ignore pycache
        if '__pycache__' in content:
            content.remove('__pycache__')
        assert sorted(content) == sorted(expected), f"Content mismatch: {sorted(content)} vs. {sorted(expected)}"

    # NOTE: contents() seems to be broken in importlib.resources (python 3.9) for zipped eggs, as it also recursively
    # lists all subdirectories (but not their files) instead of just directories within the package directory.

    # Top-level package
    expected = ['__init__.py', 'a.py', 'b.py', 'subpkg1', 'subpkg2', 'subpkg3']
    if is_frozen:
        expected.remove('subpkg2')  # FIXME: frozen reader does not list directories that are not on filesystem.
    _contents_test('pyi_pkgres_testpkg', expected)

    # Subpackage #1
    expected = ['__init__.py', 'c.py', 'd.py', 'data']
    _contents_test('pyi_pkgres_testpkg.subpkg1', expected)

    # Subpackage #2
    if not is_frozen:
        # Cannot list non-existing directory.
        expected = ['__init__.py', 'mod.py', 'subsubpkg21']
        _contents_test('pyi_pkgres_testpkg.subpkg2', expected)

    # Sub-subpackage in subpackage #2
    if not is_frozen:
        # Cannot list non-existing directory.
        expected = ['__init__.py', 'mod.py']
        _contents_test('pyi_pkgres_testpkg.subpkg2.subsubpkg21', expected)

    # Subpackage #3
    expected = ['__init__.py', '_datafile.json']
    _contents_test('pyi_pkgres_testpkg.subpkg3', expected)
else:
    print(f"Skipping {package_name}.contents() test...")

########################################################################
#              Validate importlib.resources.is_resource()              #
########################################################################
if hasattr(importlib_resources, "is_resource"):
    print(f"Testing {package_name}.is_resource()...")

    # In general, files (source or data) count as resources, directories do not.

    # Querying non-existent resource: return False instead of raising FileNotFoundError
    assert not importlib_resources.is_resource('pyi_pkgres_testpkg', 'subpkg_nonexistant')
    assert not importlib_resources.is_resource('pyi_pkgres_testpkg.subpkg1', 'nonexistant.txt')

    # NOTE: frozen reader does not list .py files (nor equivalent .pyc ones)
    assert importlib_resources.is_resource('pyi_pkgres_testpkg', '__init__.py') is not is_frozen
    assert importlib_resources.is_resource('pyi_pkgres_testpkg', '__init__.py') is not is_frozen
    assert importlib_resources.is_resource('pyi_pkgres_testpkg', 'a.py') is not is_frozen
    assert importlib_resources.is_resource('pyi_pkgres_testpkg', 'b.py') is not is_frozen

    assert not importlib_resources.is_resource('pyi_pkgres_testpkg', 'subpkg1')
    assert not importlib_resources.is_resource('pyi_pkgres_testpkg', 'subpkg2')  # Non-existent in frozen variant.
    assert not importlib_resources.is_resource('pyi_pkgres_testpkg', 'subpkg3')

    assert importlib_resources.is_resource('pyi_pkgres_testpkg.subpkg1', '__init__.py') is not is_frozen
    assert not importlib_resources.is_resource('pyi_pkgres_testpkg.subpkg1', 'data')

    # Try to specify a sub-path; should raise ValueError, unless importlib_resources >= 6.4.0.
    try:
        ret = importlib_resources.is_resource('pyi_pkgres_testpkg.subpkg1', 'data/entry1.txt')
    except ValueError:
        pass
    except Exception:
        raise
    else:
        assert new_functional_api, "Expected a ValueError!"
        # If we passed above check for importlib_resources >= 6.4.0, check the return value.
        assert ret

    if not is_frozen:
        assert importlib_resources.is_resource('pyi_pkgres_testpkg.subpkg2', 'mod.py')

    assert importlib_resources.is_resource('pyi_pkgres_testpkg.subpkg3', '_datafile.json')
else:
    print(f"Skipping {package_name}.is_resource() test...")

########################################################################
#                  Validate importlib.resources.path()                 #
########################################################################
if hasattr(importlib_resources, "path"):
    print(f"Testing {package_name}.path()...")

    # The path() function returns on-disk path. If the resource was originally on disk, direct path to it is returned.
    # Otherwise, path to a temporary file is returned. This function is probably superseded by files() and as_file(),
    # which are more flexible; for example, path() does not allow access to files in sub-directories (only files that
    # are directly within a package directory).
    def _path_test(pkgname, resource, expected_data):
        with importlib_resources.path(pkgname, resource) as pth:
            assert isinstance(pth, pathlib.Path)
            with open(pth, 'rb') as fp:
                data = fp.read()
            if expected_data is not None:
                # Split to avoid OS-specific newline discrepancies.
                assert data.splitlines() == expected_data.splitlines()

    if not is_frozen:
        expected_data = b"""from . import a, b  # noqa: F401\nfrom . import subpkg1, subpkg2, subpkg3  # noqa: F401\n"""
        _path_test('pyi_pkgres_testpkg', '__init__.py', expected_data)

    if not is_frozen:
        expected_data = b"""#\n"""
        _path_test('pyi_pkgres_testpkg.subpkg2', 'mod.py', expected_data)

    expected_data = b"""{\n  "_comment": "Data file in supbkg3."\n}\n"""
    _path_test('pyi_pkgres_testpkg.subpkg3', '_datafile.json', expected_data)

    # Try with a non-existent file; should raise FileNotFoundError.
    # NOTE: importlib.resources in python 3.9 seems to do so, but importlib_resources 5.2.2 does not...
    try:
        _path_test('pyi_pkgres_testpkg.subpkg1', 'nonexistant.txt', None)
    except FileNotFoundError:
        pass
    except Exception:
        raise
    else:
        assert not is_builtin, "Expected a FileNotFoundError!"

    # Try to specify a sub-path; should raise ValueError, unless importlib_resources >= 6.4.0.
    try:
        _path_test('pyi_pkgres_testpkg.subpkg1', 'data/entry1.txt', None)
    except ValueError:
        pass
    except Exception:
        raise
    else:
        assert new_functional_api, "Expected a ValueError!"
else:
    print(f"Skipping {package_name}.path() test...")

########################################################################
#               Validate importlib.resources.read_binary()             #
########################################################################
if hasattr(importlib_resources, "read_binary"):
    print(f"Testing {package_name}.read_binary()...")

    # Data file in pyi_pkgres_testpkg.subpkg3
    expected_data = b"""{\n  "_comment": "Data file in supbkg3."\n}\n"""
    data = importlib_resources.read_binary('pyi_pkgres_testpkg.subpkg3', '_datafile.json')
    assert data.splitlines() == expected_data.splitlines()

    # Source file in pyi_pkgres_testpkg
    if not is_frozen:
        expected_data = b"""from . import a, b  # noqa: F401\nfrom . import subpkg1, subpkg2, subpkg3  # noqa: F401\n"""
        data = importlib_resources.read_binary('pyi_pkgres_testpkg', '__init__.py')
        assert data.splitlines() == expected_data.splitlines()

    # Try with non-existent file; should raise FileNotFoundError
    try:
        importlib_resources.read_binary('pyi_pkgres_testpkg.subpkg1', 'nonexistant.txt')
    except FileNotFoundError:
        pass
    except Exception:
        raise
    else:
        assert False, "Expected a FileNotFoundError!"

    # Try to specify sub-path; should raise ValueError, unless importlib_resources >= 6.4.0.
    try:
        importlib_resources.read_binary('pyi_pkgres_testpkg.subpkg1', 'data/entry1.txt')
    except ValueError:
        pass
    except Exception:
        raise
    else:
        assert new_functional_api, "Expected a ValueError!"
else:
    print(f"Skipping {package_name}.read_binary() test...")

########################################################################
#                Validate importlib.resources.read_text()              #
########################################################################
if hasattr(importlib_resources, "read_text"):
    print(f"Testing {package_name}.read_text()...")

    # Data file in pyi_pkgres_testpkg.subpkg3
    expected_data = """{\n  "_comment": "Data file in supbkg3."\n}\n"""
    data = importlib_resources.read_text('pyi_pkgres_testpkg.subpkg3', '_datafile.json', encoding='utf8')
    assert data.splitlines() == expected_data.splitlines()

    # Source file in pyi_pkgres_testpkg
    if not is_frozen:
        expected_data = """from . import a, b  # noqa: F401\nfrom . import subpkg1, subpkg2, subpkg3  # noqa: F401\n"""
        data = importlib_resources.read_text('pyi_pkgres_testpkg', '__init__.py', encoding='utf8')
        assert data.splitlines() == expected_data.splitlines()

    # Try with non-existent file; should raise FileNotFoundError
    try:
        importlib_resources.read_text('pyi_pkgres_testpkg.subpkg1', 'nonexistant.txt', encoding='utf8')
    except FileNotFoundError:
        pass
    except Exception:
        raise
    else:
        assert False, "Expected a FileNotFoundError!"

    # Try to specify sub-path; should raise ValueError, unless importlib_resources >= 6.4.0.
    try:
        importlib_resources.read_text('pyi_pkgres_testpkg.subpkg1', 'data/entry1.txt', encoding='utf8')
    except ValueError:
        pass
    except Exception:
        raise
    else:
        assert new_functional_api, "Expected a ValueError!"
else:
    print(f"Skipping {package_name}.read_text() test...")

########################################################################
#          Validate importlib.resources.files() and as_file()          #
########################################################################
print(f"Testing {package_name}.files() and {package_name}.as_file()...")

# files() should return a Traversable (or just a plain pathlib.Path). For on-disk resources, as_file() should not be
# opening a temporary-file copy.
pkg_path = importlib_resources.files('pyi_pkgres_testpkg')
subpkg1_path = importlib_resources.files('pyi_pkgres_testpkg.subpkg1')

# assert subpkg1_path == pkg_path / 'subpkg1'  # True only for on-disk resources!

# Try to get data file in a sub-directory with subpath consisting of single string. This simulates the use in
# https://github.com/Unidata/MetPy/blob/a3424de66a44bf3a92b0dcacf4dff82ad7b86712/src/metpy/plots/wx_symbols.py#L24-L25
# that seems to trip up the back-ported importlib_resources when we do not provide our own resource reader. In that
# case, its compatibility adapter ends up triggering the copy-to-temporary-file codepath, which errors out because the
# it ends up generating a temporary file name with separator (that would require creation of intermediate directory).
# It works correctly with built-in importlib.resources in python 3.9 (even if we do not implement a resource reader).
data_path = subpkg1_path / 'data/entry1.txt'
expected_data = b"""Data entry #1 in subpkg1/data.\n"""

with importlib_resources.as_file(data_path) as file_path:
    with open(file_path, 'rb') as fp:
        data = fp.read()
assert data.splitlines() == expected_data.splitlines()

# Try to get a data file in a sub-directory via two paths. The read contents should be the same.
# If we do not provide our own resource reader, neither way of accessing works in frozen version with back-ported
# importlib_resources (``FileNotFoundError: Can't open orphan path``).
# It works correctly with built-in importlib.resources in python 3.9 (even if we do not implement a resource reader).
expected_data = b"""Data entry #2 in `subpkg1/data`.\n"""

data_path = pkg_path / 'subpkg1' / 'data' / 'entry2.md'
with importlib_resources.as_file(data_path) as file_path:
    with open(file_path, 'rb') as fp:
        data = fp.read()
assert data.splitlines() == expected_data.splitlines()

data_path = subpkg1_path / 'data' / 'entry2.md'
with importlib_resources.as_file(data_path) as file_path:
    with open(file_path, 'rb') as fp:
        data = fp.read()
assert data.splitlines() == expected_data.splitlines()

# Test that for submodules, files() returns the path to their parent package.
# See https://github.com/pyinstaller/pyinstaller/issues/8659
#
# NOTE: passing a module name to files() seems to be supported only under python >= 3.12 (stdlib) or equivalent
# importlib_resources >= 5.12. Under earlier versions, it raises `TypeError: '<name>' is not a package`.
try:
    assert importlib_resources.files('pyi_pkgres_testpkg.a') == pkg_path
except TypeError:
    pass

try:
    assert importlib_resources.files('pyi_pkgres_testpkg.subpkg1.c') == subpkg1_path
except TypeError:
    pass



# -----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
# -----------------------------------------------------------------------------
"""
This hook allows discovery and collection of PyInstaller's internal _pyi_rth_utils module that provides utility
functions for run-time hooks.

The module is implemented in 'PyInstaller/fake-modules/_pyi_rth_utils.py'.
"""

import os

from PyInstaller import PACKAGEPATH


def pre_find_module_path(api):
    module_dir = os.path.join(PACKAGEPATH, 'fake-modules')
    api.search_dirs = [module_dir]



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.Tools import ccroot, ar
from waflib.Configure import conf


@conf
def find_gcc(conf):
    cc = conf.find_program(['gcc', 'cc'], var='CC')
    conf.get_cc_version(cc, gcc=True)
    conf.env.CC_NAME = 'gcc'


@conf
def gcc_common_flags(conf):
    v = conf.env
    v.CC_SRC_F = []
    v.CC_TGT_F = ['-c', '-o']
    if not v.LINK_CC:
        v.LINK_CC = v.CC
    v.CCLNK_SRC_F = []
    v.CCLNK_TGT_F = ['-o']
    v.CPPPATH_ST = '-I%s'
    v.DEFINES_ST = '-D%s'
    v.LIB_ST = '-l%s'
    v.LIBPATH_ST = '-L%s'
    v.STLIB_ST = '-l%s'
    v.STLIBPATH_ST = '-L%s'
    v.RPATH_ST = '-Wl,-rpath,%s'
    v.SONAME_ST = '-Wl,-h,%s'
    v.SHLIB_MARKER = '-Wl,-Bdynamic'
    v.STLIB_MARKER = '-Wl,-Bstatic'
    v.cprogram_PATTERN = '%s'
    v.CFLAGS_cshlib = ['-fPIC']
    v.LINKFLAGS_cshlib = ['-shared']
    v.cshlib_PATTERN = 'lib%s.so'
    v.LINKFLAGS_cstlib = ['-Wl,-Bstatic']
    v.cstlib_PATTERN = 'lib%s.a'
    v.LINKFLAGS_MACBUNDLE = ['-bundle', '-undefined', 'dynamic_lookup']
    v.CFLAGS_MACBUNDLE = ['-fPIC']
    v.macbundle_PATTERN = '%s.bundle'


@conf
def gcc_modifier_win32(conf):
    v = conf.env
    v.cprogram_PATTERN = '%s.exe'
    v.cshlib_PATTERN = '%s.dll'
    v.implib_PATTERN = '%s.dll.a'
    v.IMPLIB_ST = '-Wl,--out-implib,%s'
    v.CFLAGS_cshlib = []
    v.append_value('LINKFLAGS', ['-Wl,--enable-auto-import'])


@conf
def gcc_modifier_cygwin(conf):
    gcc_modifier_win32(conf)
    v = conf.env
    v.cshlib_PATTERN = 'cyg%s.dll'
    v.append_value('LINKFLAGS_cshlib', ['-Wl,--enable-auto-image-base'])
    v.CFLAGS_cshlib = []


@conf
def gcc_modifier_darwin(conf):
    v = conf.env
    v.CFLAGS_cshlib = ['-fPIC']
    v.LINKFLAGS_cshlib = ['-dynamiclib']
    v.cshlib_PATTERN = 'lib%s.dylib'
    v.FRAMEWORKPATH_ST = '-F%s'
    v.FRAMEWORK_ST = ['-framework']
    v.ARCH_ST = ['-arch']
    v.LINKFLAGS_cstlib = []
    v.SHLIB_MARKER = []
    v.STLIB_MARKER = []
    v.SONAME_ST = []


@conf
def gcc_modifier_aix(conf):
    v = conf.env
    v.LINKFLAGS_cprogram = ['-Wl,-brtl']
    v.LINKFLAGS_cshlib = ['-shared', '-Wl,-brtl,-bexpfull']
    v.SHLIB_MARKER = []


@conf
def gcc_modifier_hpux(conf):
    v = conf.env
    v.SHLIB_MARKER = []
    v.STLIB_MARKER = []
    v.CFLAGS_cshlib = ['-fPIC', '-DPIC']
    v.cshlib_PATTERN = 'lib%s.sl'


@conf
def gcc_modifier_openbsd(conf):
    conf.env.SONAME_ST = []


@conf
def gcc_modifier_osf1V(conf):
    v = conf.env
    v.SHLIB_MARKER = []
    v.STLIB_MARKER = []
    v.SONAME_ST = []


@conf
def gcc_modifier_platform(conf):
    gcc_modifier_func = getattr(conf, 'gcc_modifier_' + conf.env.DEST_OS, None)
    if gcc_modifier_func:
        gcc_modifier_func()


def configure(conf):
    conf.find_gcc()
    conf.find_ar()
    conf.gcc_common_flags()
    conf.gcc_modifier_platform()
    conf.cc_load_tools()
    conf.cc_add_flags()
    conf.link_add_flags()
    conf.check_gcc_o_space()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Functional tests for PyGObject.
"""

import pytest

from PyInstaller.utils.tests import importorskip, parametrize

# Names of all "gi.repository" packages provided by PyGObject  to be tested below, typically corresponding to
# those packages hooked by PyInstaller.
_gi_repositories = [
    ('Gst', '1.0'),
    ('GLib', '2.0'),
    ('GModule', '2.0'),
    ('GObject', '2.0'),
    ('GdkPixbuf', '2.0'),
    ('Gio', '2.0'),
    ('Clutter', '1.0'),
    ('GtkClutter', '1.0'),
    ('Champlain', '0.12'),
    ('GtkChamplain', '0.12'),
]


# Test the usability of "gi.repository" packages provided by PyGObject.
@importorskip('gi.repository')
@parametrize(
    ('repository_name', 'version'),
    [pytest.param(name, version, marks=importorskip(f'gi.repository.{name}')) for name, version in _gi_repositories],
    ids=[name for name, version in _gi_repositories],  # Ensure human-readable test parameter names.
)
def test_gi_repository(pyi_builder, repository_name, version):
    """
    Test the importability of the `gi.repository` subpackage with the passed name installed with PyGObject. For example,
    `GLib`, corresponds to the `gi.repository.GLib` subpackage. Version '1.0' are for PyGObject >=1.0,
    '2.0' for PyGObject >= 2.0. Some other libraries have strange version (e.g., Champlain).
    """

    # Test the importability of this subpackage.
    pyi_builder.test_source(
        f"""
        import sys
        import pathlib

        import gi

        # Check that package/namespace can be imported
        gi.require_version('{repository_name}', '{version}')
        from gi.repository import {repository_name}
        print({repository_name})

        # Check that the typelib is, in fact, loaded from the frozen application (instead of falling back to the copy
        # that is installed on the system).
        repo = gi.Repository.get_default()
        typelib_path = repo.get_typelib_path('{repository_name}')
        print("typelib path: ", typelib_path)

        typelib_path = pathlib.Path(typelib_path).resolve()
        application_path = pathlib.Path(sys._MEIPASS).resolve()
        if application_path not in typelib_path.parents:
            raise ValueError(f"Typelib {{str(typelib_path)!r}} is not loaded from frozen application's directory!")
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2025, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# NOTE: CPython does not, in fact, guarantee reproducibility of byte-compiled .pyc modules - see [1, 2, 3].
#
# In our case, the lack of .pyc reproducibility creeps up when trying to byte-compile the same module multiple times
# *within the same python process*; the byte-compiled result of the very first attempt will differ from that of the
# second attempt, while result of all subsequent attempts will be the same as the second attempt. Note that this does
# not affect regular PyInstaller use (because each build is ran as a different process, so unless multiple targets are
# built in the same .spec file, each module is byte-compiled at most once) as much as it does our test suite (where
# multiple builds are executed within the context of the same `pytest` runner).
#
# The issues stemming from the non-deterministic behavior of FLAG_REF tracking can be, to some extent, mitigated by
# post-processing the byte-code using `marshalparser` tool [4]. However, this comes at performance cost that makes it
# unsuitable for the amount of processing done as part of our test suite. Furthermore, it does not seem to fully
# mitigate the behavior in python >= 3.13, where additional non-reproducibility aspects were introduced by interning
# of co_filename, co_name, and co_qualname [5].
#
# Therefore, to allow us to test at least some aspects of reproducibility, the tests in this file that perform multiple
# subsequent build runs and compare results between them first perform a throw-away "calibration" build and ignore its
# results. This should hopefully ensure reproducibility of the byte-compiled python code in subsequent, real build runs,
# and ensure that tests work as expected regardless of the order in which they are executed (or when executed on their
# own).
#
# References:
# [1] https://github.com/astral-sh/uv/issues/10619
# [2] https://github.com/python/cpython/issues/129724
# [3] https://github.com/python/cpython/issues/78274
# [4] https://pypi.org/project/marshalparser
# [5] https://github.com/python/cpython/commit/2ba2c142a615abbd8138d253edfe02426c386961

import dataclasses
import filecmp
import time

import pytest

from PyInstaller import compat


def test_reproducible_subsequent_builds(pyi_builder, tmp_path, monkeypatch):
    from PyInstaller.archive.readers import CArchiveReader

    # On GHA ubuntu runners, we seem to be getting occasional test failure due to differences in collected modules'
    # bytecode, in spite of the "calibration" step below. This seems to affect python < 3.11 on all platforms; most
    # often, it fails on linux with 3.9, but 3.8 and 3.10 have also been observed to fail. Since the lack of .pyc
    # bytecode reproducibility seems to be out of our control, xfail the test.
    if not compat.is_py311:
        pytest.xfail("Issues with reproducibility of collected modules' bytecode.")

    # On Windows, we need to set SOURCE_DATE_EPOCH for the build to be fully reproducible (i.e., the build timestamp
    # that is embedded in the executable)
    if compat.is_win:
        monkeypatch.setenv('SOURCE_DATE_EPOCH', f"{time.time():.0f}")

    # Two consecutive builds (preceded by a throw-away "calibration" one, as per note at the top of this file); ensure
    # their build and dist dirs are different.
    pyi_builder._dist_dir = tmp_path / 'dist-0'
    pyi_builder._build_dir = tmp_path / 'build-0'
    pyi_builder.test_script('pyi_helloworld.py')

    pyi_builder._dist_dir = tmp_path / 'dist-1'
    pyi_builder._build_dir = tmp_path / 'build-1'
    pyi_builder.test_script('pyi_helloworld.py')
    executable1 = pyi_builder._find_executables("pyi_helloworld")[0]
    print("First executable:", executable1)

    pyi_builder._dist_dir = tmp_path / 'dist-2'
    pyi_builder._build_dir = tmp_path / 'build-2'
    pyi_builder.test_script('pyi_helloworld.py')
    executable2 = pyi_builder._find_executables("pyi_helloworld")[0]
    print("Second executable:", executable2)

    pkg1 = CArchiveReader(executable1)
    pkg2 = CArchiveReader(executable2)

    # First, compare TOCs of embedded PYZ archives - so we can get detailed error if their elements differ.
    def _find_pyz(pkg):
        pyz = [name for name, (*_, typecode) in pkg.toc.items() if typecode == 'z']
        return pyz[0]

    pyz1_name = _find_pyz(pkg1)
    pyz1 = pkg1.open_embedded_archive(pyz1_name)

    pyz2_name = _find_pyz(pkg2)
    pyz2 = pkg2.open_embedded_archive(pyz2_name)

    assert pyz1.toc == pyz2.toc

    # Compare PYZ archives bit-by-bit
    pyz1_data = pkg1.extract(pyz1_name)
    pyz2_data = pkg2.extract(pyz2_name)

    assert pyz1_data == pyz2_data

    # Compare PKG TOCs
    assert pkg1.toc == pkg2.toc
    assert pkg1.options == pkg2.options

    # Compare PKG archives bit-by-bit
    pkg1_data = pkg1.raw_pkg_data()
    pkg2_data = pkg2.raw_pkg_data()

    assert pkg1_data == pkg2_data

    # Compare whole executables
    assert filecmp.cmp(executable1, executable2, shallow=False)


# Test that in macOS executables, Mach-O image UUIDs are modified during the build process, based on the contents (hash)
# of the PKG archive.
@pytest.mark.darwin
def test_macos_executable_uuid(pyi_builder_spec, tmp_path):
    # Helper for reading Mach-O image UUIDs from executable.
    def _read_uuids(filename):
        import uuid

        from macholib.mach_o import LC_UUID
        from macholib.MachO import MachO

        uuids = []

        executable = MachO(filename)
        for header in executable.headers:
            # Find LC_UUID command
            uuid_cmd = [cmd for cmd in header.commands if cmd[0].cmd == LC_UUID]
            if not uuid_cmd:
                continue
            uuid_cmd = uuid_cmd[0]

            # Copy UUID
            uuids.append(uuid.UUID(bytes=uuid_cmd[1].uuid))

        return uuids

    # Helper structure for organizing the results of build runs.
    @dataclasses.dataclass
    class Result:
        # For each resulting executable, we store  the list of UUIDs extracted from all available arch slices.
        program1_onedir: list
        program1_onefile: list
        program1_onedir_w: list
        program1_onedir_w_app: list

        program2_onedir: list
        program2_onefile: list

        program3_onedir: list
        program3_onefile: list

        @staticmethod
        def create_result(dist_path):
            return Result(
                # First program
                program1_onedir=_read_uuids(dist_path / "program1_onedir/program1_onedir"),
                program1_onefile=_read_uuids(dist_path / "program1_onefile"),
                program1_onedir_w=_read_uuids(dist_path / "program1_onedir_w/program1_onedir_w"),
                program1_onedir_w_app=_read_uuids(dist_path / "program1_onedir_w.app/Contents/MacOS/program1_onedir_w"),
                # Second program
                program2_onedir=_read_uuids(dist_path / "program2_onedir/program2_onedir"),
                program2_onefile=_read_uuids(dist_path / "program2_onefile"),
                # Third program
                program3_onedir=_read_uuids(dist_path / "program3_onedir/program3_onedir"),
                program3_onefile=_read_uuids(dist_path / "program3_onefile"),
            )

        def display_result(self, build_name):
            print(f"{build_name}")

            print("* program1_onedir:", self.program1_onedir)
            print("* program1_onefile:", self.program1_onefile)
            print("* program1_onedir_w:", self.program1_onedir_w)
            print("* program1_onedir_w_app:", self.program1_onedir_w_app)

            print("* program2_onedir:", self.program2_onedir)
            print("* program2_onedfile:", self.program2_onefile)

            print("* program3_onedir:", self.program3_onedir)
            print("* program3_onefile:", self.program3_onefile)

    # Build the test programs - throw-away "calibration" build, as per note at the top of this file.
    pyi_builder_spec._dist_dir = tmp_path / 'dist-0'
    pyi_builder_spec._build_dir = tmp_path / 'build-0'
    pyi_builder_spec.test_spec(
        'pyi_macos_executable_uuid.spec',
        # Provide name of one application to appease the find/run-executable phase.
        app_name='program1_onedir',
    )

    # Build the test programs - first run.
    pyi_builder_spec._dist_dir = tmp_path / 'dist-1'
    pyi_builder_spec._build_dir = tmp_path / 'build-1'
    pyi_builder_spec.test_spec(
        'pyi_macos_executable_uuid.spec',
        app_name='program1_onedir',
    )

    result1 = Result.create_result(pyi_builder_spec._dist_dir)
    result1.display_result("First build")

    # onefile and onedir have different contents of PKG archive.
    assert result1.program1_onedir != result1.program1_onefile

    # windowed mode uses different bootloader executable.
    assert result1.program1_onedir != result1.program1_onedir_w

    # In windowed mode, regular onedir build uses same executable as the .app bundle. So this test is technically a
    # no-op, unless we try to ensure that .app bundle codepath doesn't additionally reset the UUIDs in some way.
    assert result1.program1_onedir_w == result1.program1_onedir_w_app

    # program2 is built using the same script as program1.
    assert result1.program2_onedir == result1.program1_onedir
    assert result1.program2_onefile == result1.program1_onefile

    # program3 is built using different script.
    assert result1.program3_onedir != result1.program1_onedir
    assert result1.program3_onefile != result1.program1_onefile

    # Build the test programs - second run.
    pyi_builder_spec._dist_dir = tmp_path / 'dist-2'
    pyi_builder_spec._build_dir = tmp_path / 'build-2'
    pyi_builder_spec.test_spec(
        'pyi_macos_executable_uuid.spec',
        app_name='program1_onedir',
    )

    result2 = Result.create_result(pyi_builder_spec._dist_dir)
    result2.display_result("Second build")

    # Compare both results - the UUIDs should be identical across all pairs of corresponding programs.
    assert result1 == result2



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib.Tools import ccroot
from waflib import Utils
from waflib.Logs import debug

c_compiler = {
    'win32': ['msvc', 'gcc', 'clang'],
    'cygwin': ['gcc', 'clang'],
    'darwin': ['clang', 'gcc'],
    'aix': ['xlc', 'gcc', 'clang'],
    'linux': ['gcc', 'clang', 'icc'],
    'sunos': ['suncc', 'gcc'],
    'irix': ['gcc', 'irixcc'],
    'hpux': ['gcc'],
    'osf1V': ['gcc'],
    'gnu': ['gcc', 'clang'],
    'java': ['gcc', 'msvc', 'clang', 'icc'],
    'default': ['clang', 'gcc'],
}


def default_compilers():
    build_platform = Utils.unversioned_sys_platform()
    possible_compiler_list = c_compiler.get(build_platform, c_compiler['default'])
    return ' '.join(possible_compiler_list)


def configure(conf):
    try:
        test_for_compiler = conf.options.check_c_compiler or default_compilers()
    except AttributeError:
        conf.fatal("Add options(opt): opt.load('compiler_c')")
    for compiler in re.split('[ ,]+', test_for_compiler):
        conf.env.stash()
        conf.start_msg('Checking for %r (C compiler)' % compiler)
        try:
            conf.load(compiler)
        except conf.errors.ConfigurationError as e:
            conf.env.revert()
            conf.end_msg(False)
            debug('compiler_c: %r', e)
        else:
            if conf.env.CC:
                conf.end_msg(conf.env.get_flat('CC'))
                conf.env.COMPILER_CC = compiler
                conf.env.commit()
                break
            conf.env.revert()
            conf.end_msg(False)
    else:
        conf.fatal('could not configure a C compiler!')


def options(opt):
    test_for_compiler = default_compilers()
    opt.load_special_tools('c_*.py', ban=['c_dumbpreproc.py'])
    cc_compiler_opts = opt.add_option_group('Configuration options')
    cc_compiler_opts.add_option(
        '--check-c-compiler',
        default=None,
        help='list of C compilers to try [%s]' % test_for_compiler,
        dest="check_c_compiler"
    )
    for x in test_for_compiler.split():
        opt.load('%s' % x)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.Tools import ar, d
from waflib.Configure import conf


@conf
def find_gdc(conf):
    conf.find_program('gdc', var='D')
    out = conf.cmd_and_log(conf.env.D + ['--version'])
    if out.find("gdc") == -1:
        conf.fatal("detected compiler is not gdc")


@conf
def common_flags_gdc(conf):
    v = conf.env
    v.DFLAGS = []
    v.D_SRC_F = ['-c']
    v.D_TGT_F = '-o%s'
    v.D_LINKER = v.D
    v.DLNK_SRC_F = ''
    v.DLNK_TGT_F = '-o%s'
    v.DINC_ST = '-I%s'
    v.DSHLIB_MARKER = v.DSTLIB_MARKER = ''
    v.DSTLIB_ST = v.DSHLIB_ST = '-l%s'
    v.DSTLIBPATH_ST = v.DLIBPATH_ST = '-L%s'
    v.LINKFLAGS_dshlib = ['-shared']
    v.DHEADER_ext = '.di'
    v.DFLAGS_d_with_header = '-fintfc'
    v.D_HDR_F = '-fintfc-file=%s'


def configure(conf):
    conf.find_gdc()
    conf.load('ar')
    conf.load('d')
    conf.common_flags_gdc()
    conf.d_platform_flags()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import ctypes
import ctypes.util
import sys

import pytest

from PyInstaller.compat import is_win
from PyInstaller.utils.tests import skipif


def test_ctypes_hooks_are_in_place(pyi_builder):
    pyi_builder.test_source(
        """
        import ctypes
        assert ctypes.CDLL.__name__ == 'PyInstallerCDLL', ctypes.CDLL
        """
    )


def test_ctypes_load_and_use_dll(pyi_builder, compiled_dylib):
    pyi_builder.test_source(
        f"""
        import os
        import sys
        import ctypes

        libname = {str(compiled_dylib.name)!r}
        libpath = os.path.join(os.path.dirname(__file__), libname)

        lib = ctypes.CDLL(libpath)

        assert lib.add_twelve(42) == 42 + 12
        """,
        # Collect the compiled shared library into top-level application directory.
        pyi_args=['--add-binary', f"{compiled_dylib}:."],
    )


# Test for error message raised by our ctypes hook when the specified shared library cannot be loaded.
def test_ctypes_cdll_unknown_dll(pyi_builder, capfd):
    with pytest.raises(pytest.fail.Exception, match="Running exe .* failed"):
        pyi_builder.test_source(
            """
            import ctypes
            ctypes.cdll.LoadLibrary('non-existing-2017')
            """
        )
    out, err = capfd.readouterr()
    assert "Failed to load dynlib/dll" in err


# Make sure we are able to load CDLL(None) -> pip does this for some reason
@skipif(is_win, reason="ctypes.CDLL(None) is not valid on Windows")
def test_ctypes_cdll_none(pyi_builder):
    pyi_builder.test_source(
        """
        import ctypes
        lib = ctypes.CDLL(None)
        assert lib is not None
        """
    )


def _monkeypatch_resolveCtypesImports(monkeypatch, compiled_dylib):
    import PyInstaller.depend.utils

    _orig_resolveCtypesImports = PyInstaller.depend.utils._resolveCtypesImports

    def mocked_resolveCtypesImports(*args, **kwargs):
        from PyInstaller.config import CONF
        old_pathex = CONF['pathex']
        CONF['pathex'].append(str(compiled_dylib))
        res = _orig_resolveCtypesImports(*args, **kwargs)
        CONF['pathex'] = old_pathex
        return res

    # Add the path to ctypes_dylib to pathex, only for _resolveCtypesImports. We can not monkeypath CONF['pathex'] here,
    # as it will be overwritten when pyi_builder is starting up. So be monkeypatch _resolveCtypesImports by a wrapper.
    monkeypatch.setattr(PyInstaller.depend.utils, "_resolveCtypesImports", mocked_resolveCtypesImports)


# Check that using `ctypes.CDLL(ctypes.util.find_library(name_literal))` ends up collecting the shared library  and
# loads the bundled copy at the run-time. This test is Linux-specific and uses system-installed shared libraries. This
# is because `ctypes.util.find_library` works with linker-like names (without "lib" prefix and ".so*" suffix and
# requires the shared library to have a valid SONAME.
@pytest.mark.linux
@pytest.mark.parametrize('libname', ['png', 'gs'])
def test_ctypes_find_library_and_cdll_on_linux(pyi_builder, libname):
    # Check that the library can be resolved via ctypes.util.find_library(), which implies its availability.
    soname = ctypes.util.find_library(libname)
    if not soname:
        pytest.skip(f"could not resolve {libname} into .so library name (library is likely not installed)")

    pyi_builder.test_source(
        f"""
        import sys
        import os
        import ctypes
        import ctypes.util

        # Try to resolve the library into soname.
        # PyInstaller's ctypes analysis can pick up only literal arguments to ctypes.util.find_library()!
        soname = ctypes.util.find_library({libname!r})
        print("Resolved soname:", soname)
        assert soname is not None, 'Could not resolve linker-like library name into .so name'

        # Load the library
        lib = ctypes.CDLL(soname)
        assert lib is not None and lib._name is not None, f'Could not load shared library {{soname}}'

        # Ensure that library with the resolved soname was in fact collected into top-level application directory.
        libfile = os.path.join(sys._MEIPASS, soname)
        assert os.path.isfile(libfile), f'Shared library {{soname}} not found in top-level application directory!'

        # Check that the library we loaded is in fact the copy from the top-level application directory. For that,
        # we need psutil.
        try:
            import psutil
        except ModuleNotFoundError:
            print("psutil not available, ending test.")

        process = psutil.Process(os.getpid())
        print("Loaded libraries:")
        for loaded_lib in process.memory_maps():
            print(f"  {{loaded_lib.path}}")

        # Find the library that matches our soname
        matching_libs = [
            loaded_lib.path for loaded_lib in process.memory_maps()
            if os.path.basename(loaded_lib.path) == soname
        ]

        print(f"Matching libraries: {{matching_libs!r}}")
        assert any(loaded_lib == libfile for loaded_lib in matching_libs), f'Bundled copy of {{soname}} not loaded!'
        """
    )


#-- Generate test-cases for the different types of ctypes objects.
_ctypes_parameters = []
_ctypes_ids = []
for prefix in ('', 'ctypes.'):
    for funcname in ('CDLL', 'PyDLL', 'WinDLL', 'OleDLL', 'cdll.LoadLibrary'):
        _ctypes_ids.append(prefix + funcname)
        params = (prefix + funcname, _ctypes_ids[-1])
        if funcname in ("WinDLL", "OleDLL"):
            # WinDLL, OleDLL only work on Windows.
            params = pytest.param(*params, marks=pytest.mark.win32)
        _ctypes_parameters.append(params)


@pytest.mark.parametrize("funcname,test_id", _ctypes_parameters, ids=_ctypes_ids)
def test_ctypes_bindings(pyi_builder, monkeypatch, compiled_dylib, funcname, test_id):
    # Evaluate the soname here, so the test-code contains a constant. We want the name of the dynamically-loaded library
    # only, not its path. See discussion in https://github.com/pyinstaller/pyinstaller/pull/1478#issuecomment-139622994.
    soname = compiled_dylib.name

    # Patch _resolveCtypesImports tp extend search path with the parent directory of the compiled shared library.
    _monkeypatch_resolveCtypesImports(monkeypatch, compiled_dylib.parent)

    pyi_builder.test_source(
        f"""
        import os
        import sys

        # Both imports allow us to prefixed and non-prefixed funcname.
        import ctypes
        from ctypes import *

        # Load the library.
        lib = {funcname}({soname!r})
        print(f"Loaded library handle: {{lib}}")

        libfile = os.path.join(sys._MEIPASS, {soname!r})
        assert os.path.isfile(libfile), f'Shared library {{soname}} not found in top-level application directory!'

        # NOTE: since we are using our own compiled test library, we do not need to explicitly check that the bundled
        # copy is loaded.
        """,
        test_id=test_id,
    )


@pytest.mark.parametrize("funcname,test_id", _ctypes_parameters, ids=_ctypes_ids)
def test_ctypes_bindings_in_function(pyi_builder, monkeypatch, compiled_dylib, funcname, test_id):
    # This is much like test_ctypes_bindings except that the ctypes calls are in a function. See issue #1620.
    soname = compiled_dylib.name

    _monkeypatch_resolveCtypesImports(monkeypatch, compiled_dylib.parent)

    pyi_builder.test_source(
        f"""
        import os
        import sys

        # Both imports allow us to prefixed and non-prefixed funcname.
        import ctypes
        from ctypes import *

        # Load the library.
        def f():
            def g():
                lib = {funcname}({soname!r})
                return lib
            return g()

        lib = f()
        print(f"Loaded library handle: {{lib}}")

        libfile = os.path.join(sys._MEIPASS, {soname!r})
        assert os.path.isfile(libfile), f'Shared library {{soname}} not found in top-level application directory!'

        # NOTE: since we are using our own compiled test library, we do not need to explicitly check that the bundled
        # copy is loaded.
        """,
        test_id=test_id,
    )


def test_ctypes_cdll_builtin_extension(pyi_builder):
    # Take a built-in that is provided as an extension
    builtin_ext = '_sha256'
    if builtin_ext in sys.builtin_module_names:
        # On Windows, built-ins do not seem to be extensions
        pytest.skip(f"{builtin_ext} is a built-in module without extension.")

    pyi_builder.test_source(
        f"""
        import ctypes
        import importlib.machinery

        # Try to load CDLL with all possible extension suffices; this should fail in all cases, as built-in extensions
        # should not be in the ctypes' search path.
        builtin_ext = {builtin_ext!r}
        for suffix in importlib.machinery.EXTENSION_SUFFIXES:
            try:
                lib = ctypes.CDLL(builtin_ext + suffix)
            except OSError:
                lib = None
            assert lib is None, "Built-in extension picked up by ctypes.CDLL!"
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller import compat
from PyInstaller.utils.hooks.setuptools import setuptools_info


def pre_safe_import_module(api):
    # `distutils` was removed from from stdlib in python 3.12; if it is available, it is provided by `setuptools`.
    # Therefore, we need to create package/module alias entries, which prevent the setuptools._distutils` and its
    # submodules from being collected as top-level modules (as `distutils` and its submodules) in addition to being
    # collected as their "true" names.
    if compat.is_py312 and setuptools_info.distutils_vendored:
        for aliased_name, real_vendored_name in setuptools_info.get_distutils_aliases():
            api.add_alias_module(real_vendored_name, aliased_name)



#-----------------------------------------------------------------------------
# Copyright (c) 2020-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# This assists in creating an ``.egg`` package for use with the ``test_pkg_resources_provider``. To build the package,
# execute ``python setup.py bdist_egg``.

import setuptools

setuptools.setup(
    name='pyi_pkgres_testpkg',
    version='1.0.0',
    setup_requires="setuptools >= 40.0.0",
    author='PyInstaller development team',
    packages=setuptools.find_packages(),
    package_data={
        "pyi_pkgres_testpkg": [
            "subpkg1/data/*.txt",
            "subpkg1/data/*.md",
            "subpkg1/data/*.rst",
            "subpkg1/data/extra/*.json",
            "subpkg3/*.json",
        ],
    }
)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This contains tests for the class:``Tree``, see
# https://pyinstaller.readthedocs.io/en/latest/advanced-topics.html#the-tree-class

import os
import pytest

import PyInstaller.building.datastruct


class Tree(PyInstaller.building.datastruct.Tree):
    # A stripped-down version of PyInstaller.building.datastruct.Tree that does not check the guts,
    # but only the `assemble()` step.
    def __postinit__(self):
        self.assemble()


TEST_MOD = 'Tree_files'
_DATA_BASEPATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), TEST_MOD)

_TEST_FILES = sorted([
    os.path.join('subpkg', 'twelve.py'),
    os.path.join('subpkg', 'thirteen.txt'),
    os.path.join('subpkg', 'init__.py'),
    'two.py',
    'dynamiclib.dylib',
    os.path.join('py_files_not_in_package', 'sub_pkg', 'three.py'),
    os.path.join('py_files_not_in_package', 'sub_pkg', 'init__.py'),
    os.path.join('py_files_not_in_package', 'one.py'),
    os.path.join('py_files_not_in_package', 'data', 'eleven.dat'),
    os.path.join('py_files_not_in_package', 'ten.dat'),
    'dynamiclib.dll',
    'pyextension.pyd',
    'nine.dat',
    'init__.py',
    'pyextension.so',
])

_PARAMETERS = (
    (None, None, _TEST_FILES),
    ('abc', None, [os.path.join('abc', f) for f in _TEST_FILES]),
    (None, ['*.py'], [f for f in _TEST_FILES if not f.endswith('.py')]),
    (None, ['*.py', '*.pyd'], [f for f in _TEST_FILES if not f.endswith(('.py', '.pyd'))]),
    (None, ['subpkg'], [f for f in _TEST_FILES if not f.startswith('subpkg')]),
    (None, ['subpkg', 'sub_pkg'],
     [f for f in _TEST_FILES if not (f.startswith('subpkg') or os.sep + 'sub_pkg' + os.sep in f)]),
    ('klm', ['subpkg', 'sub_pkg', '*.py', '*.pyd'], [
        os.path.join('klm', f) for f in _TEST_FILES
        if not (f.startswith('subpkg') or os.sep + 'sub_pkg' + os.sep in f or f.endswith(('.py', '.pyd')))
    ]),
)  # yapf: disable


@pytest.mark.parametrize("prefix,excludes,result", _PARAMETERS)
def test_Tree(monkeypatch, prefix, excludes, result):
    monkeypatch.setattr('PyInstaller.config.CONF', {'workpath': '.'})
    tree = Tree(_DATA_BASEPATH, prefix=prefix, excludes=excludes)
    files = sorted(f[0] for f in tree)
    assert files == sorted(result)



import ctypes
from ctypes import wintypes
import json
import sys


class STARTUPINFOW(ctypes.Structure):
    _fields_ = [
        ("cb", wintypes.DWORD),
        ("lpReserved", wintypes.LPWSTR),
        ("lpDesktop", wintypes.LPWSTR),
        ("lpTitle", wintypes.LPWSTR),
        ("dwX", wintypes.DWORD),
        ("dwY", wintypes.DWORD),
        ("dwXSize", wintypes.DWORD),
        ("dwYSize", wintypes.DWORD),
        ("dwXCountChars", wintypes.DWORD),
        ("dwYCountChars", wintypes.DWORD),
        ("dwFillAttribute", wintypes.DWORD),
        ("dwFlags", wintypes.DWORD),
        ("wShowWindow", wintypes.WORD),
        ("cbReserved2", wintypes.WORD),
        ("lpReserved2", wintypes.LPBYTE),
        ("hStdInput", wintypes.HANDLE),
        ("hStdOutput", wintypes.HANDLE),
        ("hStdError", wintypes.HANDLE),
    ]


# Get process' startup info...
si = STARTUPINFOW()
si.cb = ctypes.sizeof(STARTUPINFOW)
ctypes.windll.kernel32.GetStartupInfoW(ctypes.byref(si))

# ... and dump fields-of-interest into JSON
FIELDS = {"dwFlags", "wShowWindow"}
result = {field: getattr(si, field) for field in FIELDS}

if len(sys.argv) > 1:
    with open(sys.argv[1], 'w') as fp:
        json.dump(result, fp)
else:
    print(json.dumps(result))
